import os
from typing import Dict, List, Tuple

import hypothesis
import hypothesis.strategies as st
import pytest
import torch

import rust_circuit as rc
from rust_circuit.py_utils import timed

from .test_rust_rewrite import get_c_st


def print_multiline_escape(s: str):
    uuid = "6c6805a0-3aaf-492c-93a3-00851eacaacf"
    print(s.replace("\n", uuid).encode("unicode_escape").decode("ASCII").replace(uuid, "\n"))


@st.composite
def st_maybe_many_circuits(draw: st.DrawFn, *args, **kwargs) -> Tuple[List[rc.Circuit], bool]:
    c_st = get_c_st(*args, **kwargs)
    many = draw(st.booleans())
    if many:
        num = draw(st.integers(min_value=0, max_value=4))
        return ([draw(c_st) for _ in range(num)], many)
    else:
        return ([draw(c_st)], many)


@hypothesis.settings(deadline=None)
@hypothesis.given(cs_is_many=st_maybe_many_circuits())
@pytest.mark.parametrize("names", [False, True])
@pytest.mark.parametrize("shapes_only_necessary", [False, True])
@pytest.mark.parametrize("force_use_serial_numbers", [False, True])
@pytest.mark.parametrize("only_child_below", [False, True])
def test_rust_printing_bijection(
    cs_is_many: Tuple[List[rc.Circuit], bool],
    names: bool,
    shapes_only_necessary: bool,
    force_use_serial_numbers: bool,
    only_child_below: bool,
):
    circs, is_many = cs_is_many
    if not names:
        circs = [rc.strip_names_and_tags(circ) for circ in circs]
    printed = rc.PrintOptions(
        True,
        shape_only_when_necessary=shapes_only_necessary,
        force_use_serial_numbers=force_use_serial_numbers,
        only_child_below=only_child_below,
    ).repr(*circs)
    hypothesis.note(printed)
    # print_multiline_escape(printed)
    parser = rc.Parser(module_check_all_inputs_used=False)
    if is_many:
        parsed = parser.parse_circuits(printed)
    else:
        parsed = [parser.parse_circuit(printed)]
    parsed = [rc.deep_canonicalize(p) for p in parsed]
    circs = [rc.deep_canonicalize(c) for c in circs]
    if parsed is not None:
        if parsed != circs:
            for p, c in zip(parsed, circs):
                c.print()
                p.print()
        assert parsed == circs


def test_empty_einsum_parse():
    assert rc.Parser()("""0 [] Einsum ->""") == rc.Einsum(out_axes=()).rename(None)


def test_autoname_disable_error():
    s = """
    0 [5, 3, 6] Add
    0 AD Add
    """
    with pytest.raises(rc.ParseCircuitOnCircuitRepeatInfoIsNotSameError):
        rc.Parser().parse_circuits(s)


def test_rust_printing_bijection_2():
    strings = [
        """
0 [5, 3, 6] Add
  1 [3, 1] Scatter [0:2,0:1]
    2 [2, 1] Scalar 1.0
  10 [5, 3, 6] Scatter [0:5,0:3,0:3]
    4 [5, 3, 3] Concat 0
      5 '\\\\'\\\\' [2, 3, 3] Scalar 1.2
      6 '\\' hi\\\\' [3, 3, 3] Scalar 1.2""",
        """
3 Tag 1a280c74-fa9f-4dd9-94f4-5cca4a0cf243
  2 [3, 3, 3] Scalar 1.2""",
        """
0 Cumulant
  1 [2, 3, 3] Scalar 1.0
  3 Tag 1a280c74-fa9f-4dd9-94f4-5cca4a0cf243
    2 [3, 3, 3] Scalar 1.2""",
        """
0 DiscreteVar
  2 [10,11] Scalar 2.2
  1 [10] Scalar 0.1
  """,
        """
0 StoredCumulantVar 1,2|1a280c74-fa9f-4dd9-94f4-5cca4a0cf243
  1 [3, 1] Scalar 1.2
  2 [3, 1, 3, 1] Scalar 1.2""",
    ]
    for s in strings:
        c = rc.Parser()(s)
        print(s)
        c.print()


def test_rust_printing_comments():
    strings = [
        """
0 [5, 3, 6] Add
  1 [3, 1] Scatter [0:2,0:1]
    2 [2, 1] Scalar 1.0
  10 [5, 3, 6] Scatter [0:5,0:3,0:3]
    4 '# my name' [5, 3, 3] Concat 0
      5 '#\\\\'\\\\' [2, 3, 3] Scalar 1.2
      6 '\\' hi\\\\' [3, 3, 3] Scalar 1.2""",
        """
        # comment before
0 [5, 3, 6] Add
  1 [3, 1] Scatter [0:2,0:1]
    2 [2, 1] Scalar 1.0
  10 [5, 3, 6] Scatter [0:5,0:3,0:3]
    4 '# my name' [5, 3, 3] Concat 0
      5 '#\\\\'\\\\' [2, 3, 3] Scalar 1.2
      6 '\\' hi\\\\' [3, 3, 3] Scalar 1.2""",
        """
# comment before
0 [5, 3, 6] Add
  1 [3, 1] Scatter [0:2,0:1]
    2 [2, 1] Scalar 1.0
  10 [5, 3, 6] Scatter [0:5,0:3,0:3]
    4 '# my name' [5, 3, 3] Concat 0
      5 '#\\\\'\\\\' [2, 3, 3] Scalar 1.2
      6 '\\' hi\\\\' [3, 3, 3] Scalar 1.2""",
        """
0 [5, 3, 6] Add
  1 [3, 1] Scatter [0:2,0:1]
# comment mid
    2 [2, 1] Scalar 1.0
  10 [5, 3, 6] Scatter [0:5,0:3,0:3]
    4 '# my name' [5, 3, 3] Concat 0
      5 '#\\\\'\\\\' [2, 3, 3] Scalar 1.2
      6 '\\' hi\\\\' [3, 3, 3] Scalar 1.2""",
        """
0 [5, 3, 6] Add
  1 [3, 1] Scatter [0:2,0:1] # comment
    2 [2, 1] Scalar 1.0
  10 [5, 3, 6] Scatter [0:5,0:3,0:3]
    4 '# my name' [5, 3, 3] Concat 0
      5 '#\\\\'\\\\' [2, 3, 3] Scalar 1.2
      6 '\\' hi\\\\' [3, 3, 3] Scalar 1.2""",
        """
0 [5, 3, 6] Add#c
  1 [3, 1] Scatter [0:2,0:1]
    2 [2, 1] Scalar 1.0
  10 [5, 3, 6] Scatter [0:5,0:3,0:3]
    4 '# my name' [5, 3, 3] Concat 0#dfj
      5 '#\\\\'\\\\' [2, 3, 3] Scalar 1.2      #  sldkj#*$&fstuf
      6 '\\' hi\\\\' [3, 3, 3] Scalar 1.2""",
        """
0 [5, 3, 6] Add
  1 [3, 1] Scatter [0:2,0:1]
    2 [2, 1] Scalar 1.0
  10 [5, 3, 6] Scatter [0:5,0:3,0:3]###
    4 '# my name' [5, 3, 3] Concat 0
      5 '#\\\\'\\\\' [2, 3, 3] Scalar 1.2
      6 '\\' hi\\\\' [3, 3, 3] Scalar 1.2""",
        """
0 [5, 3, 6] Add
  1 [3, 1] Scatter [0:2,0:1]
    # comment mid
    2 [2, 1] Scalar 1.0
  10 [5, 3, 6] Scatter [0:5,0:3,0:3]
    4 '# my name' [5, 3, 3] Concat 0
      5 '#\\\\'\\\\' [2, 3, 3] Scalar 1.2
      6 '\\' hi\\\\' [3, 3, 3] Scalar 1.2""",
    ]
    circs = [rc.Parser()(s) for s in strings]
    assert len(set(circs)) == 1, circs


def test_rust_printing_many():
    strings = [
        """
0 [5, 3, 6] Add
  1 [3, 1] Scatter [0:2,0:1]
    2 [2, 1] Scalar 1.0
  10 [5, 3, 6] Scatter [0:5,0:3,0:3]
    4 '# my name' [5, 3, 3] Concat 0
      5 '#\\\\'\\\\' [2, 3, 3] Scalar 1.2
      6 '\\' hi\\\\' [3, 3, 3] Scalar 1.2""",
        """
        # comment before
0 [5, 3, 6] Add
  1 [3, 1] Scatter [0:2,0:1]
    2 [2, 1] Scalar 1.0
  10 [5, 3, 6] Scatter [0:5,0:3,0:3]
    4 '# my name' [5, 3, 3] Concat 0
      5 '#\\\\'\\\\' [2, 3, 3] Scalar 1.2
      6 '\\' hi\\\\' [3, 3, 3] Scalar 1.2""",
        """
# comment before
0 [5, 3, 6] Add
  1 [3, 1] Scatter [0:2,0:1]
    2 [2, 1] Scalar 1.0
  10 [5, 3, 6] Scatter [0:5,0:3,0:3]
    4 '# my name' [5, 3, 3] Concat 0
      5 '#\\\\'\\\\' [2, 3, 3] Scalar 1.2
      6 '\\' hi\\\\' [3, 3, 3] Scalar 1.2""",
        """
0 [5, 3, 6] Add
  1 [3, 1] Scatter [0:2,0:1]
# comment mid
    2 [2, 1] Scalar 1.0
  10 [5, 3, 6] Scatter [0:5,0:3,0:3]
    4 '# my name' [5, 3, 3] Concat 0
      5 '#\\\\'\\\\' [2, 3, 3] Scalar 1.2
      6 '\\' hi\\\\' [3, 3, 3] Scalar 1.2""",
        """
0 [5, 3, 6] Add
  1 [3, 1] Scatter [0:2,0:1]
    # comment mid
    2 [2, 1] Scalar 1.0
  10 [5, 3, 6] Scatter [0:5,0:3,0:3]
    4 '# my name' [5, 3, 3] Concat 0
      5 '#\\\\'\\\\' [2, 3, 3] Scalar 1.2
      6 '\\' hi\\\\' [3, 3, 3] Scalar 1.2""",
    ]
    circs = [rc.Parser()(s) for s in strings]
    assert len(set(circs)) == 1, circs


def test_rust_printing_bijection_many():
    base = """
0 Add
  1 [3, 1] Scatter [0:2,0:1]
    2 [2, 1] Scalar 1
  3 '0' [1, 3, 6] Scalar 7
  4 '1' [5, 3, 6] Scatter [0:5,0:3,0:3]
    5 Concat 0
      6 [2, 3, 3] Scalar 1.2
      7 [3, 3, 3] Scalar 1.2
  8 'ein' Einsum abc->abc
    4
  3"""
    circ = rc.Parser()(base)
    extra_circs = """
# extra
3 '0' [1, 3, 6] Scalar 7
4 '1' [5, 3, 6] Scatter [0:5,0:3,0:3]
  5 Concat 0
    6 [2, 3, 3] Scalar 1.2
    7 [3, 3, 3] Scalar 1.2
# we can include newlines between

0 Add
  1 [3, 1] Scatter [0:2,0:1]
    2 [2, 1] Scalar 1
  # now below is repeated
  3
  4
  8 'ein' Einsum abc->abc
    4
  3"""

    scalar, scatter, same_circ = rc.Parser().parse_circuits(extra_circs)
    assert scalar == rc.Matcher("0").get_unique(circ)
    assert scatter == rc.Matcher("1").get_unique(circ)
    assert same_circ == circ

    expected = """
'0' [1,3,6] Scalar 7
'1' [5,3,6] Scatter [0:5,0:3,0:3]
  0 Concat 0
    1 [2,3,3] Scalar 1.2
    2 [3,3,3] Scalar 1.2
3 Add
  4 [3,1] Scatter [0:2,0:1]
    5 [2,1] Scalar 1
  '0'
  '1'
  'ein' Einsum abc->abc
    '1'
  '0'""".strip(
        "\n"
    )
    # if you update printer, verify print is as expected and then copy in new test strings
    assert expected == rc.PrintOptions().repr(scalar, scatter, same_circ)

    scalar, scatter, same_circ = rc.Parser().parse_circuits(expected)
    assert scalar == rc.Matcher("0").get_unique(circ)
    assert scatter == rc.Matcher("1").get_unique(circ)
    assert same_circ == circ


def test_rust_printing_bijection_indent():
    base = """
0 Add
  1 [3, 1] Scatter [0:2,0:1]
    2 [2, 1] Scalar 1
  3 '0' [1, 3, 6] Scalar 7
  4 '1' [5, 3, 6] Scatter [0:5,0:3,0:3]
    5 Concat 0
      6 [2, 3, 3] Scalar 1.2
      7 [3, 3, 3] Scalar 1.2
  8 'ein' Einsum abc->abc
    4
  3"""
    circ = rc.Parser()(base)
    extra_circs = """
    # extra
    3 '0' [1, 3, 6] Scalar 7
    4 '1' [5, 3, 6] Scatter [0:5,0:3,0:3]
      5 Concat 0
        6 [2, 3, 3] Scalar 1.2
        7 [3, 3, 3] Scalar 1.2
    # we can include newlines between

    0 Add
      1 [3, 1] Scatter [0:2,0:1]
        2 [2, 1] Scalar 1
      # now below is repeated
      3
      4
      8 'ein' Einsum abc->abc
        4
      3
    """

    scalar, scatter, same_circ = rc.Parser().parse_circuits(extra_circs)
    assert scalar == rc.Matcher("0").get_unique(circ)
    assert scatter == rc.Matcher("1").get_unique(circ)
    assert same_circ == circ

    extra_circs = """
        # extra indent for no reason
           # extra
           3 '0' [1, 3, 6] Scalar 7
           4 '1' [5, 3, 6] Scatter [0:5,0:3,0:3]
             5 Concat 0
               6 [2, 3, 3] Scalar 1.2
               7 [3, 3, 3] Scalar 1.2
         # we can include newlines between

           0 Add
             1 [3, 1] Scatter [0:2,0:1]
               2 [2, 1] Scalar 1
             # now below is repeated
             3
             4
             8 'ein' Einsum abc->abc
               4
             3
    """

    scalar, scatter, same_circ = rc.Parser().parse_circuits(extra_circs)
    assert scalar == rc.Matcher("0").get_unique(circ)
    assert scatter == rc.Matcher("1").get_unique(circ)
    assert same_circ == circ

    extra_circs = """
        # extra
           3 '0' [1, 3, 6] Scalar 7
           4 '1' [5, 3, 6] Scatter [0:5,0:3,0:3]
             5 Concat 0
               6 [2, 3, 3] Scalar 1.2
               7 [3, 3, 3] Scalar 1.2
         # we can include newlines between

           0 Add
             1 [3, 1] Scatter [0:2,0:1]
               2 [2, 1] Scalar 1
             # now below is repeated
             3
             4
             8 'ein' Einsum abc->abc
               4
             3"""

    scalar, scatter, same_circ = rc.Parser().parse_circuits(extra_circs)
    assert scalar == rc.Matcher("0").get_unique(circ)
    assert scatter == rc.Matcher("1").get_unique(circ)
    assert same_circ == circ

    extra_circs = """
           3 '0' [1, 3, 6] Scalar 7
           4 '1' [5, 3, 6] Scatter [0:5,0:3,0:3]
             5 Concat 0
               6 [2, 3, 3] Scalar 1.2
               7 [3, 3, 3] Scalar 1.2
           0 Add
             1 [3, 1] Scatter [0:2,0:1]
               2 [2, 1] Scalar 1
             # now below is repeated
             3
             4
             8 'ein' Einsum abc->abc
               4
             3"""

    scalar, scatter, same_circ = rc.Parser().parse_circuits(extra_circs)
    assert scalar == rc.Matcher("0").get_unique(circ)
    assert scatter == rc.Matcher("1").get_unique(circ)
    assert same_circ == circ


def test_parse_less_indent():
    less_indent = """
        8 [2] Scalar 3
    0 Add
      1 [3, 1] Scatter [0:2,0:1]
        2 [2, 1] Scalar 1
      3 '0' [1, 3, 6] Scalar 7
      4 '1' [5, 3, 6] Scatter [0:5,0:3,0:3]
        5 Concat 0
          6 [2, 3, 3] Scalar 1.2
          7 [3, 3, 3] Scalar 1.2
      8 'ein' Einsum abc->abc
        4
      3
    """
    with pytest.raises(rc.ParseCircuitLessIndentationThanFirstItemError):
        rc.Parser()(less_indent)


def test_print_tensor_index():
    array_of_tensor = rc.Array(torch.tensor([1, 2, 3, 3, 2, 1], dtype=torch.int64))
    array_of_tensor.print()
    less_indent = rc.Index(rc.Scalar(1.0, (10, 10)), (array_of_tensor.value, 0))
    assert (
        less_indent.repr()
        == """0 [6] Index [t292d9fdb22 [6],0]
  1 [10,10] Scalar 1"""
    )
    less_indent.print()
    assert P(less_indent.repr()) == less_indent


def test_parse_invalid_bang():
    incorrect_bang = """
    0 Add
      1 [3, 1] Scatter [0:2,0:1]
        2 [2, 1] Scalar 1 ! hi
      3 '0' [1, 3, 6] Scalar 7
      4 '1' [5, 3, 6] Scatter [0:5,0:3,0:3]
        5 Concat 0
          6 [2, 3, 3] Scalar 1.2
          7 [3, 3, 3] Scalar 1.2
      8 'ein' Einsum abc->abc
        4
      3
    """
    with pytest.raises(rc.ParseCircuitUnexpectedChildInfoError):
        rc.Parser()(incorrect_bang)

    incorrect_bang = """
    0 Add
      1 [3, 1] Scatter [0:2,0:1]
        2 [2, 1] Scalar 1
      3 '0' [1, 3, 6] Scalar 7
      4 '1' [5, 3, 6] Scatter [0:5,0:3,0:3]
        5 Concat 0
          6 [2, 3, 3] Scalar 1.2
          7 [3, 3, 3] Scalar 1.2
      8 'ein' Einsum abc->abc
        4 ! bye
      3
    """
    with pytest.raises(rc.ParseCircuitUnexpectedChildInfoError):
        rc.Parser()(incorrect_bang)


def test_rust_printing_repeated_names():
    stuff = """
    0 'hi' Add
      1 ' my' [3, 1] Scatter [0:2,0:1]
        2 '# 43' [2, 1] Scalar 1
      3 '1' [1, 3, 6] Scalar 7
      4 '1' [5, 3, 6] Scatter [0:5,0:3,0:3]
        5 'name stuff  ' Concat 0
          6 'hi' [2, 3, 3] Scalar 1.2
          7 [3, 3, 3] Scalar 1.2
      8 'ein' Einsum abc->abc
        4
      3
      9 'eins' Einsum ab->
        2
      10 'einy' Einsum abc->
        6
    """
    out = rc.Parser()(stuff).repr()
    expected = """
0 'hi' Add
  ' my' [3,1] Scatter [0:2,0:1]
    '# 43' [2,1] Scalar 1
  1 '1' [1,3,6] Scalar 7
  2 '1' [5,3,6] Scatter [0:5,0:3,0:3]
    'name stuff  ' Concat 0
      3 'hi' [2,3,3] Scalar 1.2
      4 [3,3,3] Scalar 1.2
  'ein' Einsum abc->abc
    2 '1'
  1 '1'
  'eins' Einsum ab->
    '# 43'
  'einy' Einsum abc->
    3 'hi'""".strip(
        "\n"
    )
    assert out == expected


def test_rust_printing_bijection_ref():
    setups: List[Tuple[str, Dict, str]] = [
        (
            """
0 [5, 3, 6] Add
  1 [3, 1] Scatter [0:2,0:1]
    'asdfadsf'
  10 [5, 3, 6] Scatter [0:5,0:3,0:3]
    4 [5, 3, 3] Concat 0
      5 'sc1' [2, 3, 3] Scalar 1.0
      6 'sc2' [3, 3, 3] Scalar 1.0""",
            {"asdfadsf": rc.Scalar(1.0, (2, 1), "asdfadsf_the_referred")},
            """
0 Add
  1 [3,1] Scatter [0:2,0:1]
    'asdfadsf_the_referred' [2,1] Scalar 1
  2 [5,3,6] Scatter [0:5,0:3,0:3]
    3 Concat 0
      'sc1' [2,3,3] Scalar 1
      'sc2' [3,3,3] Scalar 1""".strip(
                "\n"
            ),
        ),
    ]
    for string, refs, expected in setups:
        c = rc.Parser(reference_circuits=refs)(string)
        # if you update printer, verify print is as expected and then copy in new test strings
        c.print()
        print(expected)
        assert c.repr() == expected


def test_rust_printing_bijection_errors_correctly():
    strs = [
        """0 [5, 3, 6] Add d""",
        """0 [5, 3, 6] Add
 1 [3, 1] Scalar 1.0""",
        """0 [5, 3, 6] Add
    1 [3, 1] Scalar 1.0""",
        "0 [5, 3, 6] Add\n  1 [3,6] Scalar 1d",
        "0 [5, 3, 6] Add\n  1 [3,6] Index a[:3,:6]\n    2 [3,6] Scalar 1.0",
        "0 [5, 3, 6] Add\n  1 [3,6] Index [:3,:6]\n    2 [1,6] Scalar 1.0",
    ]
    for s in strs:
        try:
            c = rc.Parser()(s)
            c.print()
            print("YIKES!")
            assert False
        except ValueError as e:
            print(e)


def test_print_options():
    circ = rc.Parser()(
        """
0 [5, 3, 6] Add
  1 [3, 1] Scatter [0:2,0:1]
    2 [2, 1] Scalar 1.0
  10 [5, 3, 6] Scatter [0:5,0:3,0:3]
    4 [5, 3, 3] Concat 0
      5 'yoyo ' [2, 3, 3] Scalar 1.0
      6 'hihi' [3, 3, 3] Scalar 1.0"""
    )
    t = rc.new_traversal(end_depth=3)
    printer = rc.PrintOptions(bijection=False, traversal=t)
    # if you update printer, verify print is as expected and then copy in new test strings
    assert (
        printer.repr(circ)
        == """
0 Add
  1 [3,1] Scatter [0:2,0:1]
    2 [2,1] Scalar 1
  3 [5,3,6] Scatter [0:5,0:3,0:3]
    4 Concat 0 ...""".strip(
            "\n"
        )
    )

    assert printer.traversal == rc.new_traversal(end_depth=3)
    assert rc.PrintOptions(traversal=None).traversal == rc.new_traversal()
    assert rc.PrintHtmlOptions().overall_traversal == rc.IterativeMatcher(False)

    assert rc.PrintOptions(traversal="hi", bijection=False).print(circ) == circ.print()

    with pytest.raises(TypeError):
        rc.PrintOptions(traversal=3, bijection=False).print(circ)  # type: ignore

    printer = rc.PrintOptions(bijection=False, traversal=t, force_use_serial_numbers=True)
    # if you update printer, verify print is as expected and then copy in new test strings
    assert (
        printer.repr(circ)
        == """
0 Add
  1 [3,1] Scatter [0:2,0:1]
    2 [2,1] Scalar 1
  3 [5,3,6] Scatter [0:5,0:3,0:3]
    4 Concat 0 ...""".strip(
            "\n"
        )
    )
    assert (
        rc.PrintOptions(bijection=False, arrows=True, leaves_on_top=False).repr(circ)
        == """
0 Add
├‣1 [3,1] Scatter [0:2,0:1]
│ └‣2 [2,1] Scalar 1
└‣3 [5,3,6] Scatter [0:5,0:3,0:3]
  └‣4 Concat 0
    ├‣yoyo  [2,3,3] Scalar 1
    └‣hihi [3,3,3] Scalar 1""".strip(
            "\n"
        )
    )
    assert (
        rc.PrintOptions(bijection=False, arrows=True, leaves_on_top=True).repr(circ)
        == """
    ┌‣hihi [3,3,3] Scalar 1
    ├‣yoyo  [2,3,3] Scalar 1
  ┌‣4 Concat 0
┌‣3 [5,3,6] Scatter [0:5,0:3,0:3]
│ ┌‣2 [2,1] Scalar 1
├‣1 [3,1] Scatter [0:2,0:1]
0 Add""".strip(
            "\n"
        )
    )

    assert (
        rc.PrintOptions.repr_depth(circ, end_depth=2)
        == """0 Add
  1 [3,1] Scatter [0:2,0:1] ...
  2 [5,3,6] Scatter [0:5,0:3,0:3] ..."""
    )


def test_size_threshold_commenter():
    circ = rc.Parser()(
        """
0 [5, 3, 6] Add
  1 [3, 1] Scatter [0:2,0:1]
    2 [2, 1] Scalar 1.0
  10 [5, 3, 6] Scatter [0:5,0:3,0:3]
    4 [5, 3, 3] Concat 0
      5 'yoyo ' [2, 3, 3] Scalar 1.0
      6 'hihi' [3, 3, 3] Scalar 1.0"""
    )

    print(rc.PrintOptions(commenters=[rc.PrintOptions.size_threshold_commenter(16)]).repr(circ))

    assert (
        rc.PrintOptions(commenters=[rc.PrintOptions.size_threshold_commenter(16)]).repr(circ)
        == """0 Add # \x1b[31m90\x1b[0m
  1 [3,1] Scatter [0:2,0:1]
    2 [2,1] Scalar 1
  3 [5,3,6] Scatter [0:5,0:3,0:3] # \x1b[31m90\x1b[0m
    4 Concat 0 # \x1b[31m45\x1b[0m
      'yoyo ' [2,3,3] Scalar 1 # \x1b[31m18\x1b[0m
      'hihi' [3,3,3] Scalar 1 # \x1b[31m27\x1b[0m"""
    )


def test_dtype_commenter():
    circ = rc.Add(rc.Array(torch.zeros(5, dtype=torch.int8)), rc.Array(torch.zeros(5, dtype=torch.int8)))
    assert (
        rc.PrintOptions(commenters=[rc.PrintOptions.dtype_commenter()]).repr(circ)
        == """0 Add # int8
  1 [5] Array 50dba7b000ecd7c013961b6c # int8
  1 Array # int8"""
    )
    assert (
        rc.PrintOptions(commenters=[rc.PrintOptions.dtype_commenter(only_arrays=True)]).repr(circ)
        == """0 Add
  1 [5] Array 50dba7b000ecd7c013961b6c # int8
  1 Array # int8"""
    )
    assert (
        rc.PrintOptions(commenters=[rc.PrintOptions.dtype_commenter()]).repr(rc.Scalar(0.7, (1,)))
        == """0 [1] Scalar 0.7"""
    )


def test_default_print():
    assert "RR_DEBUG_END_DEPTH" not in os.environ
    circ = rc.Parser()(
        """
0 [5, 3, 6] Add
  1 [3, 1] Scalar 3.7
  2 [3, 1] Scatter [0:2,0:1]
    3 [2, 1] Scalar 1.0
  10 [5, 3, 6] Scatter [0:5,0:3,0:3]
    4 [5, 3, 3] Concat 0
      5 'yoyo ' [2, 3, 3] Scalar 1.0
      6 'hihi' [3, 3, 3] Scalar 1.0"""
    )
    # if you update printer, verify print is as expected and then copy in new test strings
    assert (
        str(circ)
        == """<0 [5,3,6] Add
  1 [3,1] Scalar 3.7
  2 [3,1] Scatter [0:2,0:1] ...
  3 [5,3,6] Scatter [0:5,0:3,0:3] ...>"""
    )

    try:
        rc.set_debug_print_options(rc.PrintOptions.debug_default().evolve(traversal=rc.new_traversal(end_depth=3)))
        assert (
            str(circ)
            == """<0 [5,3,6] Add
  1 [3,1] Scalar 3.7
  2 [3,1] Scatter [0:2,0:1]
    3 [2,1] Scalar 1
  4 [5,3,6] Scatter [0:5,0:3,0:3]
    5 [5,3,3] Concat 0 ...>"""
        )
    finally:
        rc.set_debug_print_options(rc.PrintOptions.debug_default())


def test_print_color():
    circ = rc.Parser()(
        """
0 Add
  1 [3, 1] Scatter [0:2,0:1]
    2 [2, 1] Scalar 1.0
  4 Concat 0
    5 'yoyo ' [2, 3, 3] Scalar 1.0
    6 'hihi' [3, 3, 3] Scalar 1.0
    7 'hihi' [3, 3, 3] Scalar 1.1
    8 'hihi' [3, 3, 3] Scalar 1.2
    9 'hihi' [3, 3, 3] Scalar 1.3
    10 'hihi' [3, 3, 3] Scalar 1.4
    11 'hihi' [3, 3, 3] Scalar 1.5
    12 'hihi' [3, 3, 3] Scalar 1.6
    13 'hihi' [3, 3, 3] Scalar 1.7
    14 'hihi' [3, 3, 3] Scalar 1.8
    15 'hihi' [3, 3, 3] Scalar 1.9
      """
    )

    colorers = [
        (
            lambda x: abs(hash(x)),
            """
\x1b[95m0 Add\x1b[0m
  \x1b[93m1 [3,1] Scatter [0:2,0:1]\x1b[0m
    \x1b[95m2 [2,1] Scalar 1\x1b[0m
  \x1b[31m3 Concat 0\x1b[0m
    \x1b[92m'yoyo ' [2,3,3] Scalar 1\x1b[0m
    \x1b[92m4 'hihi' [3,3,3] Scalar 1\x1b[0m
    \x1b[32m5 'hihi' [3,3,3] Scalar 1.1\x1b[0m
    \x1b[34m6 'hihi' [3,3,3] Scalar 1.2\x1b[0m
    \x1b[33m7 'hihi' [3,3,3] Scalar 1.3\x1b[0m
    \x1b[95m8 'hihi' [3,3,3] Scalar 1.4\x1b[0m
    \x1b[93m9 'hihi' [3,3,3] Scalar 1.5\x1b[0m
    \x1b[96m10 'hihi' [3,3,3] Scalar 1.6\x1b[0m
    \x1b[32m11 'hihi' [3,3,3] Scalar 1.7\x1b[0m
    \x1b[90m12 'hihi' [3,3,3] Scalar 1.8\x1b[0m
    \x1b[31m13 'hihi' [3,3,3] Scalar 1.9\x1b[0m""".strip(
                "\n"
            ),
        ),
        (
            rc.PrintOptions.hash_colorer(),
            """
\x1b[36m0 Add\x1b[0m
  \x1b[91m1 [3,1] Scatter [0:2,0:1]\x1b[0m
    \x1b[95m2 [2,1] Scalar 1\x1b[0m
  \x1b[33m3 Concat 0\x1b[0m
    \x1b[92m'yoyo ' [2,3,3] Scalar 1\x1b[0m
    \x1b[92m4 'hihi' [3,3,3] Scalar 1\x1b[0m
    \x1b[32m5 'hihi' [3,3,3] Scalar 1.1\x1b[0m
    \x1b[97m6 'hihi' [3,3,3] Scalar 1.2\x1b[0m
    \x1b[31m7 'hihi' [3,3,3] Scalar 1.3\x1b[0m
    \x1b[95m8 'hihi' [3,3,3] Scalar 1.4\x1b[0m
    \x1b[91m9 'hihi' [3,3,3] Scalar 1.5\x1b[0m
    \x1b[96m10 'hihi' [3,3,3] Scalar 1.6\x1b[0m
    \x1b[32m11 'hihi' [3,3,3] Scalar 1.7\x1b[0m
    \x1b[90m12 'hihi' [3,3,3] Scalar 1.8\x1b[0m
    \x1b[33m13 'hihi' [3,3,3] Scalar 1.9\x1b[0m""".strip(
                "\n"
            ),
        ),
        (
            rc.PrintOptions.type_colorer(),
            """
\x1b[96m0 Add\x1b[0m
  \x1b[33m1 [3,1] Scatter [0:2,0:1]\x1b[0m
    \x1b[35m2 [2,1] Scalar 1\x1b[0m
  \x1b[92m3 Concat 0\x1b[0m
    \x1b[35m'yoyo ' [2,3,3] Scalar 1\x1b[0m
    \x1b[35m4 'hihi' [3,3,3] Scalar 1\x1b[0m
    \x1b[35m5 'hihi' [3,3,3] Scalar 1.1\x1b[0m
    \x1b[35m6 'hihi' [3,3,3] Scalar 1.2\x1b[0m
    \x1b[35m7 'hihi' [3,3,3] Scalar 1.3\x1b[0m
    \x1b[35m8 'hihi' [3,3,3] Scalar 1.4\x1b[0m
    \x1b[35m9 'hihi' [3,3,3] Scalar 1.5\x1b[0m
    \x1b[35m10 'hihi' [3,3,3] Scalar 1.6\x1b[0m
    \x1b[35m11 'hihi' [3,3,3] Scalar 1.7\x1b[0m
    \x1b[35m12 'hihi' [3,3,3] Scalar 1.8\x1b[0m
    \x1b[35m13 'hihi' [3,3,3] Scalar 1.9\x1b[0m""".strip(
                "\n"
            ),
        ),
    ]
    print()
    for colorer, expected in colorers:
        print()
        rc.PrintOptions(colorer=colorer).print(circ)
        print()
        print_multiline_escape(rc.PrintOptions(colorer=colorer).repr(circ))
        print()
        assert rc.PrintOptions(colorer=colorer).repr(circ) == expected

    # for i in range(20):
    #     rc.PrintOptions(colorer=rc.PrintOptions.fixed_color(i)).print(rc.Scalar(1, (2, 3)))


def test_size_colorer():
    circ = rc.Parser()(
        """
0 [] Add
  1 Einsum i->
    2 [1] Scalar 1.
  3 Einsum i->
    4 [1_000] Scalar 1.
  5 Einsum i->
    6 [1_000_000] Scalar 1.
  7 Einsum i->
    8 [1_000_000_000] Scalar 1.
  9 Einsum i->
    10 [100_000_000_000] Scalar 1.
  11 Einsum i->
    12 [10_000_000_000_000] Scalar 1.
      """
    )
    expected = """
0 Add
  1 Einsum i->
    2 [1] Scalar 1
  3 Einsum i->
    4 [1000] Scalar 1
  5 Einsum i->
    \x1b[32m6 [1000000] Scalar 1\x1b[0m
  7 Einsum i->
    \x1b[33m8 [1000000000] Scalar 1\x1b[0m
  9 Einsum i->
    \x1b[31m10 [100000000000] Scalar 1\x1b[0m
  11 Einsum i->
    \x1b[31m12 [10000000000000] Scalar 1\x1b[0m""".strip(
        "\n"
    )
    colorer = rc.PrintOptions.size_colorer()
    print()
    rc.PrintOptions(colorer=colorer).print(circ)
    print()
    print_multiline_escape(rc.PrintOptions(colorer=colorer).repr(circ))
    print()
    assert rc.PrintOptions(colorer=colorer).repr(circ) == expected


@hypothesis.settings(deadline=None)
@hypothesis.given(circ=get_c_st())
def test_self_hash(circ: rc.Circuit):
    try:
        replaced = circ.map_children(lambda x: rc.Scalar(1.0, x.shape, None))
        assert rc.compute_self_hash(replaced) == rc.compute_self_hash(circ)
    except Exception as e:
        print(circ)
        raise e


P = rc.Parser()


def test_self_hash_2():
    circuits = [
        P("0 Einsum ab,ab->\n  1 [2,3] Scalar 1\n  2 [2,3] Scalar 2"),
        P("0 Einsum ad,ad->\n  1 [2,3] Scalar 1\n  2 [2,3] Scalar 2"),
        P("0 Einsum ab,ab->\n  1 [2,3] Scalar 1.1\n  2 [2,3] Scalar 2"),
        P("0 Einsum ab,ab->a\n  1 [2,3] Scalar 1\n  2 [2,3] Scalar 2"),
        P("0 Einsum ab,ab->a\n  1 [2,5] Scalar 1\n  2 [2,5] Scalar 2"),
    ]
    for circuit in circuits:
        print(rc.compute_self_hash(circuit))


def test_print_diff():
    P = rc.Parser(tensors_as_random=True, allow_hash_with_random=True)
    circuit_pairs = [
        (
            P(
                """0 Einsum ab,bc->ac
  1 [3,4] Scalar 2
  4 [4,5] Scalar 2
    """
            ),
            P(
                """0 Einsum ab,bc->ac
  1 [3,4] Scalar 3
  4 [4,5] Scalar 2
    """
            ),
        ),
        (
            P(
                """0 Einsum ab,bc->ac
  1 Einsum ab,bc->ac
    2 [2,3] Scalar 1
    3 [3,4] Scalar 1
  4 [4,5] Scalar 2
    """
            ),
            P(
                """0 Einsum ab,bc->ac
  1 Einsum ab,bc->ac
    2 [2,3] Scalar 3
    3 [3,4] Scalar 1
  4 [4,5] Scalar 2
    """
            ),
        ),
        (
            P(
                """0 Einsum ab,bc->ac
  1 Einsum ab,bc->ac
    2 [2,3] Scalar 1
    3 [3,4] Scalar 1
  4 [4,5] Add
    5 's0' [4,1] Symbol
    6 's1' [5] Symbol
    """
            ),
            P(
                """0 Einsum ab,bc->ac
  1 Einsum ab,bc->ac
    2 [2,3] Scalar 1
    3 [3,4] Scalar 2
  4 [4,5] Add
    5 's0' [4,1] Symbol
    6 's1' [1,5] Symbol
    """
            ),
        ),
        (
            P(
                """
0 Add
  1 Add
    2 Add
      'hi' [2] Symbol
      'hi' [2] Symbol
    2
  1
  'rando' [2] Symbol
  'randosame' [2] Symbol
    """
            ),
            P(
                """
0 Add
  1 Add
    2 Add
      'ho' [2] Symbol
      'ho' [2] Symbol
    2
  1
  'rando2' [2] Symbol
  'randosame' [2] Symbol
    """
            ),
        ),
        (
            P(
                """
'logit_diff_balanced_unbalanced' Add
  'logit_balanced' Einsum b,b->
    'final.n_pos0' Add
      'final.n.w.bias' [56] Array eeb9a82a41b21694aab7901b
      'final.n.y_out_pos0' Einsum b,->b
        'final.n.y_scale_pos0' Einsum b,b->b
          'final.n.z_mean_pos0' Einsum b,bc->c
            'final.inp_pos0' Add
              'tok_embeds_pos0' Einsum b,bc->c
                'tokens_pos0' Index [0,:]
                  'input_tag_logits' Tag 062c34e2-867b-4930-8ac4-bf2b16abd1e8
                    'tokens_var' DiscreteVar
                      'tokens' [1,22,5] Array 603427427c088ab4dfadc1fb
                      0 Tag 480058d7-0dd4-4105-86cc-9fa7ed06583b
                        1 [1] Scalar 1
                'tok_embed_mat' [5,56] Array 045bc7a114683c5c98190abf
              'w.pos_embeds_pos0' Index [0,:]
                'w.pos_embeds' [22,56] Array 3bf234b0ff1bec7000d09345
              'a0.out_pos0' Add
                'a0.out_pos0_head_0_sum' Einsum ab->b
                  'a0.out_pos0_head_0' Index [0:1,:]
                    'a0.out_pos0_keep_head' Einsum bc,bcd->bd
                      'a0.comb_v_pos0' Einsum ac,acd->ad
                        'a0.probs_pos0' GeneralFunction softmax
                          'a0.scores_padded_pos0' Add
                            'a0.scores_pos_mask_pos0' Einsum ac,c->ac
                              'a0.scores_not_masked_pos0' Einsum ac,adc,->ad
                                'a0.q_p_bias_pos0' Add
                                  'a0.q_pos0' Einsum abc,c->ab
                                    'a0.w.q' [2,28,56] Array d93bfa57cf2a27b30dc250d6
                                    'a0.n_pos0' Add
                                      'a0.n.w.bias' [56] Array a1090771831dccc00a5dcefa
                                      'a0.n.y_out_pos0' Einsum b,->b
                                        'a0.n.y_scale_pos0' Einsum b,b->b
                                          'a0.n.z_mean_pos0' Einsum b,bc->c
                                            'a0.inp_pos0' Add
                                              'tok_embeds_pos0'
                                              'w.pos_embeds_pos0'
                                            'a0.n.c.sub_mean' [56,56] Array 73b0b67535e8dd93d2120db9
                                          'a0.n.w.scale' [56] Array 152c532d249ccfffc16c06c4
                                        'a0.n.full_mul_pos0' GeneralFunction rsqrt
                                          'a0.n.c.var_p_eps_pos0' Add
                                            'a0.n.var_pos0' Einsum b,b,->
                                              'a0.n.z_mean_pos0'
                                              'a0.n.z_mean_pos0'
                                              'a0.n.c.recip_h_size' [] Scalar 0.017857142857142856
                                            'a0.n.eps' [] Scalar 0.00001
                                  'a0.w.q_bias_pos0' Index [:,0,:]
                                    'a0.w.q_bias' [2,1,28] Array 06ef6fea473ce03a5ff36602
                                'a0.k_p_bias' Add
                                  'a0.k' Einsum abc,dc->adb
                                    'a0.w.k' [2,28,56] Array 8917d02e3fed8bb5f5eba8b3
                                    'a0.n' Add
                                      'a0.n.w.bias'
                                      'a0.n.y_out' Einsum ab,a->ab
                                        'a0.n.y_scale' Einsum ab,b->ab
                                          'a0.n.z_mean' Einsum ab,bc->ac
                                            'a0.inp' Add
                                              'tok_embeds' Einsum ab,bc->ac
                                                'input_tag_logits'
                                                'tok_embed_mat'
                                              'w.pos_embeds'
                                            'a0.n.c.sub_mean'
                                          'a0.n.w.scale'
                                        'a0.n.full_mul' GeneralFunction rsqrt
                                          'a0.n.c.var_p_eps' Add
                                            'a0.n.var' Einsum ab,ab,->a
                                              'a0.n.z_mean'
                                              'a0.n.z_mean'
                                              'a0.n.c.recip_h_size'
                                            'a0.n.eps'
                                  'a0.w.k_bias' [2,1,28] Array 64f30e6153bfa9b2f0c5e415
                                'a0.c.div_head_size' [] Scalar 0.1889822365046136
                              'c.neg_pos_mask' Add
                                'one' [] Scalar 1
                                2 'ScalarMul' Einsum a,->a
                                  'pos_mask' Index [:,1]
                                    'input_tag_logits'
                                  3 'unnamed' [] Scalar -1
                            'ScalarMul_pos0' Einsum ac,->ac
                              'Unsqueeze_pos0' Rearrange a:22 -> () a:22
                                'pos_mask'
                              4 'unnamed' [] Scalar -10000
                        'a0.v_p_bias' Add
                          'a0.v' Einsum abc,dc->adb
                            'a0.w.v' [2,28,56] Array 6266ba0d9297e3401ecaa15a
                            'a0.n'
                          'a0.w.v_bias' [2,1,28] Array 5480b6d7d30b03372375f556
                      'a0.w.out' [2,28,56] Array f6abfd1f4f8100780784d1f7
                'a0.out_pos0_head_1_sum' Einsum ab->b
                  'a0.out_pos0_head_1' Index [1:2,:]
                    'a0.out_pos0_keep_head'
              'a0.w.bias_out' [56] Array ef811930aeadb47e844242f0
              'm0_pos0' Einsum b,cb->c
                'm0.act_pos0' GeneralFunction relu
                  'm0.add0_pos0' Add
                    'm0.before_product0_pos0' Einsum b,cb->c
                      'm0.n_pos0' Add
                        'm0.n.w.bias' [56] Array abdfc87093d25769cb5340d8
                        'm0.n.y_out_pos0' Einsum b,->b
                          'm0.n.y_scale_pos0' Einsum b,b->b
                            'm0.n.z_mean_pos0' Einsum b,bc->c
                              'm0.inp_pos0' Add
                                'tok_embeds_pos0'
                                'w.pos_embeds_pos0'
                                'a0.out_pos0'
                                'a0.w.bias_out'
                              'm0.n.c.sub_mean' [56,56] Array 73b0b67535e8dd93d2120db9
                            'm0.n.w.scale' [56] Array eacc41edc6480f2c11a896d2
                          'm0.n.full_mul_pos0' GeneralFunction rsqrt
                            'm0.n.c.var_p_eps_pos0' Add
                              'm0.n.var_pos0' Einsum b,b,->
                                'm0.n.z_mean_pos0'
                                'm0.n.z_mean_pos0'
                                'm0.n.c.recip_h_size' [] Scalar 0.017857142857142856
                              'm0.n.eps' [] Scalar 0.00001
                      'm0.w.w0' [56,56] Array 8e3d05e33535d2ff7a650701
                    'm0.w.b0' [56] Array e533e9975b83858743c9085e
                'm0.w.w1' [56,56] Array 3c486e70d7cec981912fb94c
              'm0.w.b1' [56] Array 75aed749c4c165b60488cd7c
              'a1.out_pos0' Add
                'a1.out_pos0_head_0_sum' Einsum ab->b
                  'a1.out_pos0_head_0' Index [0:1,:]
                    'a1.out_pos0_keep_head' Einsum bc,bcd->bd
                      'a1.comb_v_pos0' Einsum ac,acd->ad
                        'a1.probs_pos0' GeneralFunction softmax
                          'a1.scores_padded_pos0' Add
                            'a1.scores_pos_mask_pos0' Einsum ac,c->ac
                              'a1.scores_not_masked_pos0' Einsum ac,adc,->ad
                                'a1.q_p_bias_pos0' Add
                                  'a1.q_pos0' Einsum abc,c->ab
                                    'a1.w.q' [2,28,56] Array d171720d20b7b5143362d3ac
                                    'a1.n_pos0' Add
                                      'a1.n.w.bias' [56] Array 9c5ca27d53f3f58ec59ec0a0
                                      'a1.n.y_out_pos0' Einsum b,->b
                                        'a1.n.y_scale_pos0' Einsum b,b->b
                                          'a1.n.z_mean_pos0' Einsum b,bc->c
                                            'a1.inp_pos0' Add
                                              'tok_embeds_pos0'
                                              'w.pos_embeds_pos0'
                                              'a0.out_pos0'
                                              'a0.w.bias_out'
                                              'm0_pos0'
                                              'm0.w.b1'
                                            'a1.n.c.sub_mean' [56,56] Array 73b0b67535e8dd93d2120db9
                                          'a1.n.w.scale' [56] Array 35f1b4f1ec7b3f7543667e09
                                        'a1.n.full_mul_pos0' GeneralFunction rsqrt
                                          'a1.n.c.var_p_eps_pos0' Add
                                            'a1.n.var_pos0' Einsum b,b,->
                                              'a1.n.z_mean_pos0'
                                              'a1.n.z_mean_pos0'
                                              'a1.n.c.recip_h_size' [] Scalar 0.017857142857142856
                                            'a1.n.eps' [] Scalar 0.00001
                                  'a1.w.q_bias_pos0' Index [:,0,:]
                                    'a1.w.q_bias' [2,1,28] Array 02a0c59e812c211740fb27bb
                                'a1.k_p_bias' Add
                                  'a1.k' Einsum abc,dc->adb
                                    'a1.w.k' [2,28,56] Array 6a260810e3b7af22a8544a16
                                    'a1.n' Add
                                      'a1.n.w.bias'
                                      'a1.n.y_out' Einsum ab,a->ab
                                        'a1.n.y_scale' Einsum ab,b->ab
                                          'a1.n.z_mean' Einsum ab,bc->ac
                                            'a1.inp' Add
                                              'tok_embeds'
                                              'w.pos_embeds'
                                              'a0.out' Einsum abc,bcd->ad
                                                'a0.comb_v' Einsum abc,acd->bad
                                                  'a0.probs' GeneralFunction softmax
                                                    'a0.scores_padded' Add
                                                      'a0.scores_pos_mask' Einsum abc,c->abc
                                                        'a0.scores_not_masked' Einsum abc,adc,->abd
                                                          'a0.q_p_bias' Add
                                                            'a0.q' Einsum abc,dc->adb
                                                              'a0.w.q'
                                                              'a0.n'
                                                            'a0.w.q_bias'
                                                          'a0.k_p_bias'
                                                          'a0.c.div_head_size'
                                                        'c.neg_pos_mask'
                                                      5 'ScalarMul' Einsum abc,->abc
                                                        'Unsqueeze' Rearrange a:22 -> () () a:22
                                                          'pos_mask'
                                                        4 'unnamed'
                                                  'a0.v_p_bias'
                                                'a0.w.out'
                                              'a0.w.bias_out'
                                              'm0' Einsum ab,cb->ac
                                                'm0.act' GeneralFunction relu
                                                  'm0.add0' Add
                                                    'm0.before_product0' Einsum ab,cb->ac
                                                      'm0.n' Add
                                                        'm0.n.w.bias'
                                                        'm0.n.y_out' Einsum ab,a->ab
                                                          'm0.n.y_scale' Einsum ab,b->ab
                                                            'm0.n.z_mean' Einsum ab,bc->ac
                                                              'm0.inp' Add
                                                                'tok_embeds'
                                                                'w.pos_embeds'
                                                                'a0.out'
                                                                'a0.w.bias_out'
                                                              'm0.n.c.sub_mean'
                                                            'm0.n.w.scale'
                                                          'm0.n.full_mul' GeneralFunction rsqrt
                                                            'm0.n.c.var_p_eps' Add
                                                              'm0.n.var' Einsum ab,ab,->a
                                                                'm0.n.z_mean'
                                                                'm0.n.z_mean'
                                                                'm0.n.c.recip_h_size'
                                                              'm0.n.eps'
                                                      'm0.w.w0'
                                                    'm0.w.b0'
                                                'm0.w.w1'
                                              'm0.w.b1'
                                            'a1.n.c.sub_mean'
                                          'a1.n.w.scale'
                                        'a1.n.full_mul' GeneralFunction rsqrt
                                          'a1.n.c.var_p_eps' Add
                                            'a1.n.var' Einsum ab,ab,->a
                                              'a1.n.z_mean'
                                              'a1.n.z_mean'
                                              'a1.n.c.recip_h_size'
                                            'a1.n.eps'
                                  'a1.w.k_bias' [2,1,28] Array 4f504eb0607ce4368fea6e28
                                'a1.c.div_head_size' [] Scalar 0.1889822365046136
                              'c.neg_pos_mask'
                            'ScalarMul_pos0'
                        'a1.v_p_bias' Add
                          'a1.v' Einsum abc,dc->adb
                            'a1.w.v' [2,28,56] Array 247f61024c75b09c2c1eb7c3
                            'a1.n'
                          'a1.w.v_bias' [2,1,28] Array 6166fa008805ef7f07a08185
                      'a1.w.out' [2,28,56] Array ba174133f8cea9ec3dbd0ad3
                'a1.out_pos0_head_1_sum' Einsum ab->b
                  'a1.out_pos0_head_1' Index [1:2,:]
                    'a1.out_pos0_keep_head'
              'a1.w.bias_out' [56] Array bf309aff1cd4964da4cf8f46
              'm1_pos0' Einsum b,cb->c
                'm1.act_pos0' GeneralFunction relu
                  'm1.add0_pos0' Add
                    'm1.before_product0_pos0' Einsum b,cb->c
                      'm1.n_pos0' Add
                        'm1.n.w.bias' [56] Array a866530bc942a654a3dfede7
                        'm1.n.y_out_pos0' Einsum b,->b
                          'm1.n.y_scale_pos0' Einsum b,b->b
                            'm1.n.z_mean_pos0' Einsum b,bc->c
                              'm1.inp_pos0' Add
                                'tok_embeds_pos0'
                                'w.pos_embeds_pos0'
                                'a0.out_pos0'
                                'a0.w.bias_out'
                                'm0_pos0'
                                'm0.w.b1'
                                'a1.out_pos0'
                                'a1.w.bias_out'
                              'm1.n.c.sub_mean' [56,56] Array 73b0b67535e8dd93d2120db9
                            'm1.n.w.scale' [56] Array f9953fb52e69501d0c0d087b
                          'm1.n.full_mul_pos0' GeneralFunction rsqrt
                            'm1.n.c.var_p_eps_pos0' Add
                              'm1.n.var_pos0' Einsum b,b,->
                                'm1.n.z_mean_pos0'
                                'm1.n.z_mean_pos0'
                                'm1.n.c.recip_h_size' [] Scalar 0.017857142857142856
                              'm1.n.eps' [] Scalar 0.00001
                      'm1.w.w0' [56,56] Array 5816774b9181cb66fd8bb246
                    'm1.w.b0' [56] Array 597a51357ba300db411e6d75
                'm1.w.w1' [56,56] Array 8ddff693c2cdc7e11833a6d7
              'm1.w.b1' [56] Array 74f91880b278f8ab6ec762e4
              'a2.out_pos0' Add
                'a2.out_pos0_head_0_sum' Einsum ab->b
                  'a2.out_pos0_head_0' Index [0:1,:]
                    'a2.out_pos0_keep_head' Einsum bc,bcd->bd
                      'a2.comb_v_pos0' Einsum ac,acd->ad
                        'a2.probs_pos0' GeneralFunction softmax
                          'a2.scores_padded_pos0' Add
                            'a2.scores_pos_mask_pos0' Einsum ac,c->ac
                              'a2.scores_not_masked_pos0' Einsum ac,adc,->ad
                                'a2.q_p_bias_pos0' Add
                                  'a2.q_pos0' Einsum abc,c->ab
                                    'a2.w.q' [2,28,56] Array 9bbbb8061e5f16ca9e30a660
                                    'a2.n_pos0' Add
                                      'a2.n.w.bias' [56] Array efc80eea18b8832c38eae67d
                                      'a2.n.y_out_pos0' Einsum b,->b
                                        'a2.n.y_scale_pos0' Einsum b,b->b
                                          'a2.n.z_mean_pos0' Einsum b,bc->c
                                            'a2.inp_pos0' Add
                                              'tok_embeds_pos0'
                                              'w.pos_embeds_pos0'
                                              'a0.out_pos0'
                                              'a0.w.bias_out'
                                              'm0_pos0'
                                              'm0.w.b1'
                                              'a1.out_pos0'
                                              'a1.w.bias_out'
                                              'm1_pos0'
                                              'm1.w.b1'
                                            'a2.n.c.sub_mean' [56,56] Array 73b0b67535e8dd93d2120db9
                                          'a2.n.w.scale' [56] Array 0ca9ee9df54aeaf7244dc23d
                                        'a2.n.full_mul_pos0' GeneralFunction rsqrt
                                          'a2.n.c.var_p_eps_pos0' Add
                                            'a2.n.var_pos0' Einsum b,b,->
                                              'a2.n.z_mean_pos0'
                                              'a2.n.z_mean_pos0'
                                              'a2.n.c.recip_h_size' [] Scalar 0.017857142857142856
                                            'a2.n.eps' [] Scalar 0.00001
                                  'a2.w.q_bias_pos0' Index [:,0,:]
                                    'a2.w.q_bias' [2,1,28] Array eec5e06f978cc768053299b6
                                'a2.k_p_bias' Add
                                  'a2.k' Einsum abc,dc->adb
                                    'a2.w.k' [2,28,56] Array 65a03f81065243b62fa9bfec
                                    'a2.n' Add
                                      'a2.n.w.bias'
                                      'a2.n.y_out' Einsum ab,a->ab
                                        'a2.n.y_scale' Einsum ab,b->ab
                                          'a2.n.z_mean' Einsum ab,bc->ac
                                            'a2.inp' Add
                                              'tok_embeds'
                                              'w.pos_embeds'
                                              'a0.out'
                                              'a0.w.bias_out'
                                              'm0'
                                              'm0.w.b1'
                                              'a1.out' Einsum abc,bcd->ad
                                                'a1.comb_v' Einsum abc,acd->bad
                                                  'a1.probs' GeneralFunction softmax
                                                    'a1.scores_padded' Add
                                                      'a1.scores_pos_mask' Einsum abc,c->abc
                                                        'a1.scores_not_masked' Einsum abc,adc,->abd
                                                          'a1.q_p_bias' Add
                                                            'a1.q' Einsum abc,dc->adb
                                                              'a1.w.q'
                                                              'a1.n'
                                                            'a1.w.q_bias'
                                                          'a1.k_p_bias'
                                                          'a1.c.div_head_size'
                                                        'c.neg_pos_mask'
                                                      5 'ScalarMul'
                                                  'a1.v_p_bias'
                                                'a1.w.out'
                                              'a1.w.bias_out'
                                              'm1' Einsum ab,cb->ac
                                                'm1.act' GeneralFunction relu
                                                  'm1.add0' Add
                                                    'm1.before_product0' Einsum ab,cb->ac
                                                      'm1.n' Add
                                                        'm1.n.w.bias'
                                                        'm1.n.y_out' Einsum ab,a->ab
                                                          'm1.n.y_scale' Einsum ab,b->ab
                                                            'm1.n.z_mean' Einsum ab,bc->ac
                                                              'm1.inp' Add
                                                                'tok_embeds'
                                                                'w.pos_embeds'
                                                                'a0.out'
                                                                'a0.w.bias_out'
                                                                'm0'
                                                                'm0.w.b1'
                                                                'a1.out'
                                                                'a1.w.bias_out'
                                                              'm1.n.c.sub_mean'
                                                            'm1.n.w.scale'
                                                          'm1.n.full_mul' GeneralFunction rsqrt
                                                            'm1.n.c.var_p_eps' Add
                                                              'm1.n.var' Einsum ab,ab,->a
                                                                'm1.n.z_mean'
                                                                'm1.n.z_mean'
                                                                'm1.n.c.recip_h_size'
                                                              'm1.n.eps'
                                                      'm1.w.w0'
                                                    'm1.w.b0'
                                                'm1.w.w1'
                                              'm1.w.b1'
                                            'a2.n.c.sub_mean'
                                          'a2.n.w.scale'
                                        'a2.n.full_mul' GeneralFunction rsqrt
                                          'a2.n.c.var_p_eps' Add
                                            'a2.n.var' Einsum ab,ab,->a
                                              'a2.n.z_mean'
                                              'a2.n.z_mean'
                                              'a2.n.c.recip_h_size'
                                            'a2.n.eps'
                                  'a2.w.k_bias' [2,1,28] Array 9381312ae69eb63bc827a201
                                'a2.c.div_head_size' [] Scalar 0.1889822365046136
                              'c.neg_pos_mask'
                            'ScalarMul_pos0'
                        'a2.v_p_bias' Add
                          'a2.v' Einsum abc,dc->adb
                            'a2.w.v' [2,28,56] Array b3ee0ca518631de9db93012c
                            'a2.n'
                          'a2.w.v_bias' [2,1,28] Array fee1b26e3976991fb36fc784
                      'a2.w.out' [2,28,56] Array a6b80e015342412235a0dfe1
                'a2.out_pos0_head_1_sum' Einsum ab->b
                  'a2.out_pos0_head_1' Index [1:2,:]
                    'a2.out_pos0_keep_head'
              'a2.w.bias_out' [56] Array 5aab3d97625d2724cd18c894
              'm2_pos0' Einsum b,cb->c
                'm2.act_pos0' GeneralFunction relu
                  'm2.add0_pos0' Add
                    'm2.before_product0_pos0' Einsum b,cb->c
                      'm2.n_pos0' Add
                        'm2.n.w.bias' [56] Array 4928c790a030f627a207af3d
                        'm2.n.y_out_pos0' Einsum b,->b
                          'm2.n.y_scale_pos0' Einsum b,b->b
                            'm2.n.z_mean_pos0' Einsum b,bc->c
                              'm2.inp_pos0' Add
                                'tok_embeds_pos0'
                                'w.pos_embeds_pos0'
                                'a0.out_pos0'
                                'a0.w.bias_out'
                                'm0_pos0'
                                'm0.w.b1'
                                'a1.out_pos0'
                                'a1.w.bias_out'
                                'm1_pos0'
                                'm1.w.b1'
                                'a2.out_pos0'
                                'a2.w.bias_out'
                              'm2.n.c.sub_mean' [56,56] Array 73b0b67535e8dd93d2120db9
                            'm2.n.w.scale' [56] Array 407fac965b99d87edd348b62
                          'm2.n.full_mul_pos0' GeneralFunction rsqrt
                            'm2.n.c.var_p_eps_pos0' Add
                              'm2.n.var_pos0' Einsum b,b,->
                                'm2.n.z_mean_pos0'
                                'm2.n.z_mean_pos0'
                                'm2.n.c.recip_h_size' [] Scalar 0.017857142857142856
                              'm2.n.eps' [] Scalar 0.00001
                      'm2.w.w0' [56,56] Array b481ef72e75d0cd36cca8f96
                    'm2.w.b0' [56] Array 48618f201fe9087f87885333
                'm2.w.w1' [56,56] Array f87c96393e560aae63e9aecd
              'm2.w.b1' [56] Array a1abdf91f20e023361a01060
            'final.n.c.sub_mean' [56,56] Array 73b0b67535e8dd93d2120db9
          'final.n.w.scale' [56] Array 4576df694468664e4ac7ef98
        'final.n.full_mul_pos0' GeneralFunction rsqrt
          'final.n.c.var_p_eps_pos0' Add
            'final.n.var_pos0' Einsum b,b,->
              'final.n.z_mean_pos0'
              'final.n.z_mean_pos0'
              'final.n.c.recip_h_size' [] Scalar 0.017857142857142856
            'final.n.eps' [] Scalar 0.00001
    'w.unembed_bal' Index [1,:]
      'w.unembed' [2,56] Array c2f2e68b7f3fd72f821721a6
  'neg logit_unbalanced' Einsum ,->
    'logit_unbalanced' Einsum b,b->
      'final.n_pos0'
      'w.unembed_unbal' Index [0,:]
        'w.unembed'
    'neg_1' [] Scalar -1
        """
            ),
            P(
                """
'logit_diff_balanced_unbalanced' Add
  'logit_balanced' Einsum b,b->
    'final.n_pos0' Add
      'final.n.w.bias' [56] Array 6b49d14d528b372702a2e3b7
      'final.n.y_out_pos0' Einsum b,->b
        'final.n.y_scale_pos0' Einsum b,b->b
          'final.n.z_mean_pos0' Einsum b,bc->c
            'final.inp_pos0' Add
              'tok_embeds_pos0' Einsum b,bc->c
                'tokens_pos0' Index [0,:]
                  'tokens' [22,5] Symbol d0ce964d-4e8e-46ce-8fa1-91fa1b7f7b92
                'tok_embed_mat' [5,56] Array 24b7cbef931e7cf5f5fc28fa
              'w.pos_embeds_pos0' Index [0,:]
                'w.pos_embeds' [22,56] Array 24e659e5f4fa92a567ab4774
              'a0.out_pos0' Add
                'a0.out_pos0_head_0_sum' Einsum ab->b
                  'a0.out_pos0_head_0' Index [0:1,:]
                    'a0.out_pos0_keep_head' Einsum bc,bcd->bd
                      'a0.comb_v_pos0' Einsum ac,acd->ad
                        'a0.probs_pos0' GeneralFunction softmax
                          'a0.scores_padded_pos0' Add
                            'a0.scores_pos_mask_pos0' Einsum ac,c->ac
                              'a0.scores_not_masked_pos0' Einsum ac,adc,->ad
                                'a0.q_p_bias_pos0' Add
                                  'a0.q_pos0' Einsum abc,c->ab
                                    'a0.w.q' [2,28,56] Array 3077e7db78308fcdb8615772
                                    'a0.n_pos0' Add
                                      'a0.n.w.bias' [56] Array a2f364532da26d9ac3526286
                                      'a0.n.y_out_pos0' Einsum b,->b
                                        'a0.n.y_scale_pos0' Einsum b,b->b
                                          'a0.n.z_mean_pos0' Einsum b,bc->c
                                            'a0.inp_pos0' Add
                                              'tok_embeds_pos0'
                                              'w.pos_embeds_pos0'
                                            'a0.n.c.sub_mean' [56,56] Array 325ba9e9e24767e94536f106
                                          'a0.n.w.scale' [56] Array 8468c210313dea1608d6378d
                                        'a0.n.full_mul_pos0' GeneralFunction rsqrt
                                          'a0.n.c.var_p_eps_pos0' Add
                                            'a0.n.var_pos0' Einsum b,b,->
                                              'a0.n.z_mean_pos0'
                                              'a0.n.z_mean_pos0'
                                              'a0.n.c.recip_h_size' [] Scalar 0.017857142857142856
                                            'a0.n.eps' [] Scalar 0.00001
                                  'a0.w.q_bias_pos0' Index [:,0,:]
                                    'a0.w.q_bias' [2,1,28] Array 6428063697a3febc94f260ed
                                'a0.k_p_bias' Add
                                  'a0.k' Einsum abc,dc->adb
                                    'a0.w.k' [2,28,56] Array a3c22b26f43f450dd4417892
                                    'a0.n' Add
                                      'a0.n.w.bias'
                                      'a0.n.y_out' Einsum ab,a->ab
                                        'a0.n.y_scale' Einsum ab,b->ab
                                          'a0.n.z_mean' Einsum ab,bc->ac
                                            'a0.inp' Add
                                              'tok_embeds' Einsum ab,bc->ac
                                                'tokens'
                                                'tok_embed_mat'
                                              'w.pos_embeds'
                                            'a0.n.c.sub_mean'
                                          'a0.n.w.scale'
                                        'a0.n.full_mul' GeneralFunction rsqrt
                                          'a0.n.c.var_p_eps' Add
                                            'a0.n.var' Einsum ab,ab,->a
                                              'a0.n.z_mean'
                                              'a0.n.z_mean'
                                              'a0.n.c.recip_h_size'
                                            'a0.n.eps'
                                  'a0.w.k_bias' [2,1,28] Array d702254edc06572691386e2a
                                'a0.c.div_head_size' [] Scalar 0.1889822365046136
                              'c.neg_pos_mask' Add
                                'one' [] Scalar 1
                                0 'ScalarMul' Einsum a,->a
                                  'pos_mask' Index [:,1]
                                    'tokens'
                                  1 'unnamed' [] Scalar -1
                            'ScalarMul_pos0' Einsum ac,->ac
                              'Unsqueeze_pos0' Rearrange a:22 -> () a:22
                                'pos_mask'
                              2 'unnamed' [] Scalar -10000
                        'a0.v_p_bias' Add
                          'a0.v' Einsum abc,dc->adb
                            'a0.w.v' [2,28,56] Array 99b1866cf05e772b6645832c
                            'a0.n'
                          'a0.w.v_bias' [2,1,28] Array 8b2f6908b683001dae532509
                      'a0.w.out' [2,28,56] Array fa5cc32c773a90d37646f354
                'a0.out_pos0_head_1_sum' Einsum ab->b
                  'a0.out_pos0_head_1' Index [1:2,:]
                    'a0.out_pos0_keep_head'
              'a0.w.bias_out' [56] Array a6a4e8ab956ae6af221daddf
              'm0_pos0' Einsum b,cb->c
                'm0.act_pos0' GeneralFunction relu
                  'm0.add0_pos0' Add
                    'm0.before_product0_pos0' Einsum b,cb->c
                      'm0.n_pos0' Add
                        'm0.n.w.bias' [56] Array fec0ef838ad56e332ad9958f
                        'm0.n.y_out_pos0' Einsum b,->b
                          'm0.n.y_scale_pos0' Einsum b,b->b
                            'm0.n.z_mean_pos0' Einsum b,bc->c
                              'm0.inp_pos0' Add
                                'tok_embeds_pos0'
                                'w.pos_embeds_pos0'
                                'a0.out_pos0'
                                'a0.w.bias_out'
                              'm0.n.c.sub_mean' [56,56] Array 325ba9e9e24767e94536f106
                            'm0.n.w.scale' [56] Array 932215c27175655d1c5b3447
                          'm0.n.full_mul_pos0' GeneralFunction rsqrt
                            'm0.n.c.var_p_eps_pos0' Add
                              'm0.n.var_pos0' Einsum b,b,->
                                'm0.n.z_mean_pos0'
                                'm0.n.z_mean_pos0'
                                'm0.n.c.recip_h_size' [] Scalar 0.017857142857142856
                              'm0.n.eps' [] Scalar 0.00001
                      'm0.w.w0' [56,56] Array 2775bbed88fc828aeb2d6d81
                    'm0.w.b0' [56] Array d3370780192530e1c174dd54
                'm0.w.w1' [56,56] Array 66d491452d9c628ef37ef6b3
              'm0.w.b1' [56] Array 8985e96b24913782f3cda1fb
              'a1.out_pos0' Add
                'a1.out_pos0_head_0_sum' Einsum ab->b
                  'a1.out_pos0_head_0' Index [0:1,:]
                    'a1.out_pos0_keep_head' Einsum bc,bcd->bd
                      'a1.comb_v_pos0' Einsum ac,acd->ad
                        'a1.probs_pos0' GeneralFunction softmax
                          'a1.scores_padded_pos0' Add
                            'a1.scores_pos_mask_pos0' Einsum ac,c->ac
                              'a1.scores_not_masked_pos0' Einsum ac,adc,->ad
                                'a1.q_p_bias_pos0' Add
                                  'a1.q_pos0' Einsum abc,c->ab
                                    'a1.w.q' [2,28,56] Array 27ded626381bd8ddcbe43e24
                                    'a1.n_pos0' Add
                                      'a1.n.w.bias' [56] Array bec75fb5a4118f2a4da60794
                                      'a1.n.y_out_pos0' Einsum b,->b
                                        'a1.n.y_scale_pos0' Einsum b,b->b
                                          'a1.n.z_mean_pos0' Einsum b,bc->c
                                            'a1.inp_pos0' Add
                                              'tok_embeds_pos0'
                                              'w.pos_embeds_pos0'
                                              'a0.out_pos0'
                                              'a0.w.bias_out'
                                              'm0_pos0'
                                              'm0.w.b1'
                                            'a1.n.c.sub_mean' [56,56] Array 325ba9e9e24767e94536f106
                                          'a1.n.w.scale' [56] Array faba3366e8b1301362c9838a
                                        'a1.n.full_mul_pos0' GeneralFunction rsqrt
                                          'a1.n.c.var_p_eps_pos0' Add
                                            'a1.n.var_pos0' Einsum b,b,->
                                              'a1.n.z_mean_pos0'
                                              'a1.n.z_mean_pos0'
                                              'a1.n.c.recip_h_size' [] Scalar 0.017857142857142856
                                            'a1.n.eps' [] Scalar 0.00001
                                  'a1.w.q_bias_pos0' Index [:,0,:]
                                    'a1.w.q_bias' [2,1,28] Array e6a99b1ee0ce98064e2b5810
                                'a1.k_p_bias' Add
                                  'a1.k' Einsum abc,dc->adb
                                    'a1.w.k' [2,28,56] Array 2447e42db2e9ca421c933a86
                                    'a1.n' Add
                                      'a1.n.w.bias'
                                      'a1.n.y_out' Einsum ab,a->ab
                                        'a1.n.y_scale' Einsum ab,b->ab
                                          'a1.n.z_mean' Einsum ab,bc->ac
                                            'a1.inp' Add
                                              'tok_embeds'
                                              'w.pos_embeds'
                                              'a0.out' Einsum abc,bcd->ad
                                                'a0.comb_v' Einsum abc,acd->bad
                                                  'a0.probs' GeneralFunction softmax
                                                    'a0.scores_padded' Add
                                                      'a0.scores_pos_mask' Einsum abc,c->abc
                                                        'a0.scores_not_masked' Einsum abc,adc,->abd
                                                          'a0.q_p_bias' Add
                                                            'a0.q' Einsum abc,dc->adb
                                                              'a0.w.q'
                                                              'a0.n'
                                                            'a0.w.q_bias'
                                                          'a0.k_p_bias'
                                                          'a0.c.div_head_size'
                                                        'c.neg_pos_mask'
                                                      3 'ScalarMul' Einsum abc,->abc
                                                        'Unsqueeze' Rearrange a:22 -> () () a:22
                                                          'pos_mask'
                                                        2 'unnamed'
                                                  'a0.v_p_bias'
                                                'a0.w.out'
                                              'a0.w.bias_out'
                                              'm0' Einsum ab,cb->ac
                                                'm0.act' GeneralFunction relu
                                                  'm0.add0' Add
                                                    'm0.before_product0' Einsum ab,cb->ac
                                                      'm0.n' Add
                                                        'm0.n.w.bias'
                                                        'm0.n.y_out' Einsum ab,a->ab
                                                          'm0.n.y_scale' Einsum ab,b->ab
                                                            'm0.n.z_mean' Einsum ab,bc->ac
                                                              'm0.inp' Add
                                                                'tok_embeds'
                                                                'w.pos_embeds'
                                                                'a0.out'
                                                                'a0.w.bias_out'
                                                              'm0.n.c.sub_mean'
                                                            'm0.n.w.scale'
                                                          'm0.n.full_mul' GeneralFunction rsqrt
                                                            'm0.n.c.var_p_eps' Add
                                                              'm0.n.var' Einsum ab,ab,->a
                                                                'm0.n.z_mean'
                                                                'm0.n.z_mean'
                                                                'm0.n.c.recip_h_size'
                                                              'm0.n.eps'
                                                      'm0.w.w0'
                                                    'm0.w.b0'
                                                'm0.w.w1'
                                              'm0.w.b1'
                                            'a1.n.c.sub_mean'
                                          'a1.n.w.scale'
                                        'a1.n.full_mul' GeneralFunction rsqrt
                                          'a1.n.c.var_p_eps' Add
                                            'a1.n.var' Einsum ab,ab,->a
                                              'a1.n.z_mean'
                                              'a1.n.z_mean'
                                              'a1.n.c.recip_h_size'
                                            'a1.n.eps'
                                  'a1.w.k_bias' [2,1,28] Array a49b17b7ef53c430d7729666
                                'a1.c.div_head_size' [] Scalar 0.1889822365046136
                              'c.neg_pos_mask'
                            'ScalarMul_pos0'
                        'a1.v_p_bias' Add
                          'a1.v' Einsum abc,dc->adb
                            'a1.w.v' [2,28,56] Array 2b799f4714d3e4946f6ad692
                            'a1.n'
                          'a1.w.v_bias' [2,1,28] Array dd5f6dc596ebd588c223c68e
                      'a1.w.out' [2,28,56] Array bb36cea278ac830b8c38bf56
                'a1.out_pos0_head_1_sum' Einsum ab->b
                  'a1.out_pos0_head_1' Index [1:2,:]
                    'a1.out_pos0_keep_head'
              'a1.w.bias_out' [56] Array 6ff98054537a894d4f864eaf
              'm1_pos0' Einsum b,cb->c
                'm1.act_pos0' GeneralFunction relu
                  'm1.add0_pos0' Add
                    'm1.before_product0_pos0' Einsum b,cb->c
                      'm1.n_pos0' Add
                        'm1.n.w.bias' [56] Array e1427cbf98e903020419c988
                        'm1.n.y_out_pos0' Einsum b,->b
                          'm1.n.y_scale_pos0' Einsum b,b->b
                            'm1.n.z_mean_pos0' Einsum b,bc->c
                              'm1.inp_pos0' Add
                                'tok_embeds_pos0'
                                'w.pos_embeds_pos0'
                                'a0.out_pos0'
                                'a0.w.bias_out'
                                'm0_pos0'
                                'm0.w.b1'
                                'a1.out_pos0'
                                'a1.w.bias_out'
                              'm1.n.c.sub_mean' [56,56] Array 325ba9e9e24767e94536f106
                            'm1.n.w.scale' [56] Array a6d0f045744c39c4cc7e433e
                          'm1.n.full_mul_pos0' GeneralFunction rsqrt
                            'm1.n.c.var_p_eps_pos0' Add
                              'm1.n.var_pos0' Einsum b,b,->
                                'm1.n.z_mean_pos0'
                                'm1.n.z_mean_pos0'
                                'm1.n.c.recip_h_size' [] Scalar 0.017857142857142856
                              'm1.n.eps' [] Scalar 0.00001
                      'm1.w.w0' [56,56] Array 03029605fcf142ed26b4d7c1
                    'm1.w.b0' [56] Array bc840ad268d5ce20d2945c42
                'm1.w.w1' [56,56] Array 8fd43e4886842ad370b0d014
              'm1.w.b1' [56] Array 437b46dac5b85ba3a25ad5d2
              'a2.out_pos0' Add
                'a2.out_pos0_head_0_sum' Einsum ab->b
                  'a2.out_pos0_head_0' Index [0:1,:]
                    'a2.out_pos0_keep_head' Einsum bc,bcd->bd
                      'a2.comb_v_pos0' Einsum ac,acd->ad
                        'a2.probs_pos0' GeneralFunction softmax
                          'a2.scores_padded_pos0' Add
                            'a2.scores_pos_mask_pos0' Einsum ac,c->ac
                              'a2.scores_not_masked_pos0' Einsum ac,adc,->ad
                                'a2.q_p_bias_pos0' Add
                                  'a2.q_pos0' Einsum abc,c->ab
                                    'a2.w.q' [2,28,56] Array de2488f8a3e053ace1d73be6
                                    'a2.n_pos0' Add
                                      'a2.n.w.bias' [56] Array 8b88f4fbcf0be7097423b7c6
                                      'a2.n.y_out_pos0' Einsum b,->b
                                        'a2.n.y_scale_pos0' Einsum b,b->b
                                          'a2.n.z_mean_pos0' Einsum b,bc->c
                                            'a2.inp_pos0' Add
                                              'tok_embeds_pos0'
                                              'w.pos_embeds_pos0'
                                              'a0.out_pos0'
                                              'a0.w.bias_out'
                                              'm0_pos0'
                                              'm0.w.b1'
                                              'a1.out_pos0'
                                              'a1.w.bias_out'
                                              'm1_pos0'
                                              'm1.w.b1'
                                            'a2.n.c.sub_mean' [56,56] Array 325ba9e9e24767e94536f106
                                          'a2.n.w.scale' [56] Array 0db2d344a8397242c3d84925
                                        'a2.n.full_mul_pos0' GeneralFunction rsqrt
                                          'a2.n.c.var_p_eps_pos0' Add
                                            'a2.n.var_pos0' Einsum b,b,->
                                              'a2.n.z_mean_pos0'
                                              'a2.n.z_mean_pos0'
                                              'a2.n.c.recip_h_size' [] Scalar 0.017857142857142856
                                            'a2.n.eps' [] Scalar 0.00001
                                  'a2.w.q_bias_pos0' Index [:,0,:]
                                    'a2.w.q_bias' [2,1,28] Array a7c83d68605a729408f7680e
                                'a2.k_p_bias' Add
                                  'a2.k' Einsum abc,dc->adb
                                    'a2.w.k' [2,28,56] Array 5dbfea26b320218793d7bf39
                                    'a2.n' Add
                                      'a2.n.w.bias'
                                      'a2.n.y_out' Einsum ab,a->ab
                                        'a2.n.y_scale' Einsum ab,b->ab
                                          'a2.n.z_mean' Einsum ab,bc->ac
                                            'a2.inp' Add
                                              'tok_embeds'
                                              'w.pos_embeds'
                                              'a0.out'
                                              'a0.w.bias_out'
                                              'm0'
                                              'm0.w.b1'
                                              'a1.out' Einsum abc,bcd->ad
                                                'a1.comb_v' Einsum abc,acd->bad
                                                  'a1.probs' GeneralFunction softmax
                                                    'a1.scores_padded' Add
                                                      'a1.scores_pos_mask' Einsum abc,c->abc
                                                        'a1.scores_not_masked' Einsum abc,adc,->abd
                                                          'a1.q_p_bias' Add
                                                            'a1.q' Einsum abc,dc->adb
                                                              'a1.w.q'
                                                              'a1.n'
                                                            'a1.w.q_bias'
                                                          'a1.k_p_bias'
                                                          'a1.c.div_head_size'
                                                        'c.neg_pos_mask'
                                                      3 'ScalarMul'
                                                  'a1.v_p_bias'
                                                'a1.w.out'
                                              'a1.w.bias_out'
                                              'm1' Einsum ab,cb->ac
                                                'm1.act' GeneralFunction relu
                                                  'm1.add0' Add
                                                    'm1.before_product0' Einsum ab,cb->ac
                                                      'm1.n' Add
                                                        'm1.n.w.bias'
                                                        'm1.n.y_out' Einsum ab,a->ab
                                                          'm1.n.y_scale' Einsum ab,b->ab
                                                            'm1.n.z_mean' Einsum ab,bc->ac
                                                              'm1.inp' Add
                                                                'tok_embeds'
                                                                'w.pos_embeds'
                                                                'a0.out'
                                                                'a0.w.bias_out'
                                                                'm0'
                                                                'm0.w.b1'
                                                                'a1.out'
                                                                'a1.w.bias_out'
                                                              'm1.n.c.sub_mean'
                                                            'm1.n.w.scale'
                                                          'm1.n.full_mul' GeneralFunction rsqrt
                                                            'm1.n.c.var_p_eps' Add
                                                              'm1.n.var' Einsum ab,ab,->a
                                                                'm1.n.z_mean'
                                                                'm1.n.z_mean'
                                                                'm1.n.c.recip_h_size'
                                                              'm1.n.eps'
                                                      'm1.w.w0'
                                                    'm1.w.b0'
                                                'm1.w.w1'
                                              'm1.w.b1'
                                            'a2.n.c.sub_mean'
                                          'a2.n.w.scale'
                                        'a2.n.full_mul' GeneralFunction rsqrt
                                          'a2.n.c.var_p_eps' Add
                                            'a2.n.var' Einsum ab,ab,->a
                                              'a2.n.z_mean'
                                              'a2.n.z_mean'
                                              'a2.n.c.recip_h_size'
                                            'a2.n.eps'
                                  'a2.w.k_bias' [2,1,28] Array 4e50fd8efe268ed0ee7d4dc5
                                'a2.c.div_head_size' [] Scalar 0.1889822365046136
                              'c.neg_pos_mask'
                            'ScalarMul_pos0'
                        'a2.v_p_bias' Add
                          'a2.v' Einsum abc,dc->adb
                            'a2.w.v' [2,28,56] Array d722b26fc7e19bfececef65a
                            'a2.n'
                          'a2.w.v_bias' [2,1,28] Array 4e19a39f1496d3e67b64cad6
                      'a2.w.out' [2,28,56] Array 682c186f4fe413780c0b1e5e
                'a2.out_pos0_head_1_sum' Einsum ab->b
                  'a2.out_pos0_head_1' Index [1:2,:]
                    'a2.out_pos0_keep_head'
              'a2.w.bias_out' [56] Array f12556b8e170ce3ddde4d0d6
              'm2_pos0' Einsum b,cb->c
                'm2.act_pos0' GeneralFunction relu
                  'm2.add0_pos0' Add
                    'm2.before_product0_pos0' Einsum b,cb->c
                      'm2.n_pos0' Add
                        'm2.n.w.bias' [56] Array e2712658f0e7604ad69ae281
                        'm2.n.y_out_pos0' Einsum b,->b
                          'm2.n.y_scale_pos0' Einsum b,b->b
                            'm2.n.z_mean_pos0' Einsum b,bc->c
                              'm2.inp_pos0' Add
                                'tok_embeds_pos0'
                                'w.pos_embeds_pos0'
                                'a0.out_pos0'
                                'a0.w.bias_out'
                                'm0_pos0'
                                'm0.w.b1'
                                'a1.out_pos0'
                                'a1.w.bias_out'
                                'm1_pos0'
                                'm1.w.b1'
                                'a2.out_pos0'
                                'a2.w.bias_out'
                              'm2.n.c.sub_mean' [56,56] Array 325ba9e9e24767e94536f106
                            'm2.n.w.scale' [56] Array a4318f7b7d6eca0873e4b1c4
                          'm2.n.full_mul_pos0' GeneralFunction rsqrt
                            'm2.n.c.var_p_eps_pos0' Add
                              'm2.n.var_pos0' Einsum b,b,->
                                'm2.n.z_mean_pos0'
                                'm2.n.z_mean_pos0'
                                'm2.n.c.recip_h_size' [] Scalar 0.017857142857142856
                              'm2.n.eps' [] Scalar 0.00001
                      'm2.w.w0' [56,56] Array 378702f5015024bc11c99923
                    'm2.w.b0' [56] Array e5a37c6d6b6b95eb3363aaf9
                'm2.w.w1' [56,56] Array 6ff87833d69056a4f145c5e3
              'm2.w.b1' [56] Array 85abdbba0615b7f2574319e3
            'final.n.c.sub_mean' [56,56] Array 325ba9e9e24767e94536f106
          'final.n.w.scale' [56] Array 49283af11ca865adb6a8bd12
        'final.n.full_mul_pos0' GeneralFunction rsqrt
          'final.n.c.var_p_eps_pos0' Add
            'final.n.var_pos0' Einsum b,b,->
              'final.n.z_mean_pos0'
              'final.n.z_mean_pos0'
              'final.n.c.recip_h_size' [] Scalar 0.017857142857142856
            'final.n.eps' [] Scalar 0.00001
    'w.unembed_bal' Index [1,:]
      'w.unembed' [2,56] Array bdff3a8cee2d823672ce6624
  'neg logit_unbalanced' Einsum ,->
    'logit_unbalanced' Einsum b,b->
      'final.n_pos0'
      'w.unembed_unbal' Index [0,:]
        'w.unembed'
    'neg_1' [] Scalar -1
        """
            ),
        ),
    ]
    print()
    for n, o in circuit_pairs:
        with timed("diffing"):
            print(rc.diff_circuits(n, o, rc.PrintOptions(bijection=False), same_color="Cyan"))


def test_rust_parsing_id_by_name():
    strings = [
        """
        0 'hi' Add
          1 ' my' [3, 1] Scatter [0:2,0:1]
            2 '# 43' [2, 1] Scalar 1
          3 '0' [1, 3, 6] Scalar 7
          4 '1' [5, 3, 6] Scatter [0:5,0:3,0:3]
            5 'name stuff  ' Concat 0
              6 '###$' [2, 3, 3] Scalar 1.2
              7 [3, 3, 3] Scalar 1.2
          8 'ein' Einsum abc->abc
            4
          3
          9 'eins' Einsum ab->
            2
          10 'einy' Einsum abc->
            6
        """,
        """
        'hi' Add
          ' my' [3, 1] Scatter [0:2,0:1]
            '# 43' [2, 1] Scalar 1
          '0' [1, 3, 6] Scalar 7
          '1' [5, 3, 6] Scatter [0:5,0:3,0:3]
            'name stuff  ' Concat 0
              '###$' [2, 3, 3] Scalar 1.2
              7 [3, 3, 3] Scalar 1.2
          'ein' Einsum abc->abc
            '1'
          '0'
          'eins' Einsum ab->
            '# 43'
          'einy' Einsum abc->
            '###$'
        """,
        """
        0 'hi' Add
          1 ' my' [3, 1] Scatter [0:2,0:1]
            '    totally different' '# 43' [2, 1] Scalar 1
          3 '0' [1, 3, 6] Scalar 7
          4 '1' [5, 3, 6] Scatter [0:5,0:3,0:3]
            5 'name stuff  ' Concat 0
              '###$' [2, 3, 3] Scalar 1.2
              7 [3, 3, 3] Scalar 1.2
          8 'ein' Einsum abc->abc
            4
          3
          9 'eins' Einsum ab->
            '    totally different'
          10 'einy' Einsum abc->
            '###$'
        """,
    ]

    circs = [rc.Parser()(s) for s in strings]
    assert len(set(circs)) == 1, circs


def test_print_seen_same_line():
    circ = P(
        """
0 Add
  1 [2] Scalar 1
  2 Add
    1
    1
    3 [2] Scalar 2
    4 Add
      1
      3
    """
    )

    rc.PrintOptions(seen_children_same_line=True, bijection=False).print(circ)
    rc.PrintOptions(seen_children_same_line=True, bijection=False, arrows=True).print(circ)


def test_print_child_below():
    circ = P(
        """
0 Add
  1 Add
    3 [2] Scalar 1
  2 Add
    3
    4 [2] Scalar 1
  """
    )
    rc.PrintOptions(only_child_below=True, bijection=False).print(circ)


def test_print_traversal_by_child():
    # shamelessly taken from nest.py

    # first, we'll create a flat einsum we want to rearrange
    a, b, c, d, f, e = 2, 3, 4, 5, 6, 7
    x = rc.Einsum.from_einsum_string(
        "a b c, b c d, a e, b d, f c, a c d, e a f -> a c d f",
        *(
            rc.Array.randn(*shape, name=name)
            for (shape, name) in [
                ((a, b, c), "abc"),
                ((b, c, d), "bcd"),
                ((a, e), "ae"),
                ((b, d), "bd"),
                ((f, c), "ec"),
                ((a, c, d), "acd"),
                ((e, a, f), "eaf"),
            ]
        ),
    ).normalize_ints()
    z_new = rc.nest_einsums(x, (("acd", ("bcd", rc.Matcher.regex("^bd$"))), ("ae", "ec"), rc.NestRest()))

    fancy_sub_matcher = rc.IterativeMatcher.new_func(
        lambda _: rc.IterateMatchResults(
            [
                rc.new_traversal(term_early_at=rc.Matcher.regex(r"^\w+ \* \w+$")),
                rc.FINISHED,
                True,
            ],
            True,
        )
    )
    assert (
        rc.PrintOptions(bijection=False, traversal=fancy_sub_matcher).repr(z_new)
        == """
abc * bcd * ae * bd * ec * acd * eaf Einsum acef,abcd,abcde->acfb
  acd * bcd * bd Einsum acf,cef->acef
    acd [2,4,5] Array
    bcd * bd Einsum ecf,ef->cef ...
  ae * ec Einsum ad,bc->abcd ...
  abc * eaf Einsum aec,dab->abcde
    abc [2,3,4] Array
    eaf [7,2,6] Array""".strip(
            "\n"
        )
    )


mlp_2_layer_str = """
'outer_b.bilinear_mlp' Module
  'bilinear_mlp' Einsum h,h,oh->o
    'bilinear_mlp.pre0' Index [0,:]
      'bilinear_mlp.fold_pre' Rearrange (a:2 b) -> a:2 b
        'bilinear_mlp.pre' Einsum i,hi->h
          'bilinear_mlp.input' [?] Symbol 873a937e-2bb9-4f7f-b55e-a100db3dde52
          'bilinear_mlp.w.proj_in' [?,?] Symbol c171d519-8793-4a8b-ac5e-d550347f30a6
    'bilinear_mlp.pre1' Index [1,:]
      'bilinear_mlp.fold_pre'
    'bilinear_mlp.w.proj_out' [?,?] Symbol e61637eb-9f17-4325-b2c2-5eb2518026cf
  'b.bilinear_mlp' Module ! 'bilinear_mlp.input'
    'bilinear_mlp'
    'b.input' [3,5,7,17] Scalar 7 ! 'bilinear_mlp.input'
    'b.proj_in' [22,17] Scalar 8 ! 'bilinear_mlp.w.proj_in' ftt
    'b.proj_out' [13,11] Scalar 9 ! 'bilinear_mlp.w.proj_out' ftt
  'b_new.proj_in' [22,13] Scalar 10 ! 'bilinear_mlp.w.proj_in' ftt
  'b_new.proj_out' [20,11] Scalar 11 ! 'bilinear_mlp.w.proj_out' ftt
"""
mlp_2_layer = rc.Parser(tensors_as_random=True)(mlp_2_layer_str)


def test_rust_module_printing():
    r = mlp_2_layer.repr()
    assert rc.Parser()(r) == mlp_2_layer

    def update_mod(x: rc.Circuit):
        mod = x.cast_module()
        snd_spec = mod.spec.arg_specs[1]
        new_spec = rc.ModuleSpec(
            mod.spec.circuit,
            [
                mod.spec.arg_specs[0],
                rc.ModuleArgSpec(snd_spec.symbol, batchable=False, expandable=True, ban_non_symbolic_size_expand=False),
                *mod.spec.arg_specs[2:],
            ],
        )
        return rc.Module.new_flat(new_spec, *mod.nodes, name=mod.name)

    new_mlp_2_layer = rc.IterativeMatcher("b.bilinear_mlp").update(mlp_2_layer, update_mod)
    assert rc.Parser()(new_mlp_2_layer.repr()) == new_mlp_2_layer
    new_mlp_2_layer.print()


def test_print_comments():
    rc.PrintOptions(comment_arg_names=True).print(mlp_2_layer)
    rc.PrintOptions(comment_arg_names=True, arrows=True).print(mlp_2_layer)
    assert (
        rc.PrintOptions(comment_arg_names=True, arrows=True).repr(mlp_2_layer)
        == """
'outer_b.bilinear_mlp' Module
├‣'bilinear_mlp' Einsum h,h,oh->o # Spec
│ ├‣'bilinear_mlp.pre0' [?] Index [0,:] # h
│ │ └‣'bilinear_mlp.fold_pre' Rearrange (a:2 b) -> a:2 b
│ │   └‣'bilinear_mlp.pre' Einsum i,hi->h
│ │     ├‣'bilinear_mlp.input' [?] Symbol 873a937e-2bb9-4f7f-b55e-a100db3dde52 # i
│ │     └‣'bilinear_mlp.w.proj_in' [?,?] Symbol c171d519-8793-4a8b-ac5e-d550347f30a6 # hi
│ ├‣'bilinear_mlp.pre1' [?] Index [1,:] # h
│ │ └‣'bilinear_mlp.fold_pre'
│ └‣'bilinear_mlp.w.proj_out' [?,?] Symbol e61637eb-9f17-4325-b2c2-5eb2518026cf # oh
├‣'b.bilinear_mlp' Module ! 'bilinear_mlp.input'
│ ├‣'bilinear_mlp' # Spec
│ ├‣'b.input' [3,5,7,17] Scalar 7 ! 'bilinear_mlp.input'
│ ├‣'b.proj_in' [22,17] Scalar 8 ! 'bilinear_mlp.w.proj_in' ftt
│ └‣'b.proj_out' [13,11] Scalar 9 ! 'bilinear_mlp.w.proj_out' ftt
├‣'b_new.proj_in' [22,13] Scalar 10 ! 'bilinear_mlp.w.proj_in' ftt
└‣'b_new.proj_out' [20,11] Scalar 11 ! 'bilinear_mlp.w.proj_out' ftt
""".strip(
            "\n"
        )
    )
    rc.PrintOptions(comment_arg_names=True).print(
        rc.Parser()(
            """
0 'hoo' StoredCumulantVar 1,2,4|18af4b4c-e297-4d86-a51f-0e523d809289
  1 'hii' [] Scalar 1
  1
  1
  """
        )
    )


def test_commenters():
    commenter = lambda x: f"line length: {(len(rc.PrintOptions().repr_line_info(x)))}"
    commenter2 = lambda x: f"node count: {rc.count_nodes(x)}"
    print()
    rc.PrintOptions(commenters=[commenter, commenter2]).print(mlp_2_layer)
    assert (
        rc.PrintOptions(commenters=[commenter, commenter2]).repr(mlp_2_layer)
        == """
'outer_b.bilinear_mlp' Module # line length: 7 # node count: 15
  'bilinear_mlp' Einsum h,h,oh->o # line length: 17 # node count: 8
    'bilinear_mlp.pre0' [?] Index [0,:] # line length: 16 # node count: 5
      'bilinear_mlp.fold_pre' Rearrange (a:2 b) -> a:2 b # line length: 27 # node count: 4
        'bilinear_mlp.pre' Einsum i,hi->h # line length: 15 # node count: 3
          'bilinear_mlp.input' [?] Symbol 873a937e-2bb9-4f7f-b55e-a100db3dde52 # line length: 48 # node count: 1
          'bilinear_mlp.w.proj_in' [?,?] Symbol c171d519-8793-4a8b-ac5e-d550347f30a6 # line length: 50 # node count: 1
    'bilinear_mlp.pre1' [?] Index [1,:] # line length: 16 # node count: 5
      'bilinear_mlp.fold_pre' # line length: 27 # node count: 4
    'bilinear_mlp.w.proj_out' [?,?] Symbol e61637eb-9f17-4325-b2c2-5eb2518026cf # line length: 50 # node count: 1
  'b.bilinear_mlp' Module ! 'bilinear_mlp.input' # line length: 7 # node count: 12
    'bilinear_mlp' # line length: 17 # node count: 8
    'b.input' [3,5,7,17] Scalar 7 ! 'bilinear_mlp.input' # line length: 20 # node count: 1
    'b.proj_in' [22,17] Scalar 8 ! 'bilinear_mlp.w.proj_in' ftt # line length: 17 # node count: 1
    'b.proj_out' [13,11] Scalar 9 ! 'bilinear_mlp.w.proj_out' ftt # line length: 17 # node count: 1
  'b_new.proj_in' [22,13] Scalar 10 ! 'bilinear_mlp.w.proj_in' ftt # line length: 18 # node count: 1
  'b_new.proj_out' [20,11] Scalar 11 ! 'bilinear_mlp.w.proj_out' ftt # line length: 18 # node count: 1
""".strip(
            "\n"
        )
    )

    parsed_comment = rc.Parser()(rc.PrintOptions(commenters=[commenter, commenter2]).repr(mlp_2_layer))
    assert parsed_comment == mlp_2_layer

    # term early case
    print()
    opt = rc.PrintOptions(commenters=[commenter, commenter2], traversal=rc.new_traversal(end_depth=3), bijection=False)
    opt.print(mlp_2_layer)
    assert (
        opt.repr(mlp_2_layer)
        == """
outer_b.bilinear_mlp Module # line length: 7 # node count: 15
  bilinear_mlp Einsum h,h,oh->o # line length: 17 # node count: 8
    bilinear_mlp.pre0 [?] Index [0,:] ... # line length: 16 # node count: 5
    bilinear_mlp.pre1 [?] Index [1,:] ... # line length: 16 # node count: 5
    bilinear_mlp.w.proj_out [?,?] Symbol e61637eb-9f17-4325-b2c2-5eb2518026cf # line length: 50 # node count: 1
  b.bilinear_mlp Module ! bilinear_mlp.input [?] Symbol 873a937e-2bb9-4f7f-b55e-a100db3dde52 # line length: 7 # node count: 12
    bilinear_mlp # line length: 17 # node count: 8
    b.input [3,5,7,17] Scalar 7 ! bilinear_mlp.input # line length: 20 # node count: 1
    b.proj_in [22,17] Scalar 8 ! bilinear_mlp.w.proj_in [?,?] Symbol c171d519-8793-4a8b-ac5e-d550347f30a6 ftt # line length: 17 # node count: 1
    b.proj_out [13,11] Scalar 9 ! bilinear_mlp.w.proj_out ftt # line length: 17 # node count: 1
  b_new.proj_in [22,13] Scalar 10 ! bilinear_mlp.w.proj_in ftt # line length: 18 # node count: 1
  b_new.proj_out [20,11] Scalar 11 ! bilinear_mlp.w.proj_out ftt # line length: 18 # node count: 1
""".strip(
            "\n"
        )
    )


def test_parse_fancy():
    fancy_einsum = """
0 Einsum fancy: first second, first third -> first
  1 [3, 1] Scalar 1.0
  2 [3, 1] Scalar 1.5"""
    einsum = """
0 Einsum ab, ac -> a
  1 [3, 1] Scalar 1.0
  2 [3, 1] Scalar 1.5"""
    assert rc.Parser()(fancy_einsum) == rc.Parser()(einsum)
    fancy_einsum = """
0 Einsum fancy:first second -> second
  1 [3, 1] Scalar 1.0"""
    einsum = """
0 Einsum ab -> b
  1 [3, 1] Scalar 1.0"""
    assert rc.Parser()(fancy_einsum) == rc.Parser()(einsum)


def test_number_leaves():
    c = rc.Scalar(1.0)
    many = rc.Add(rc.Add(c, c), c, rc.Add(c, c))
    opt = rc.PrintOptions(number_leaves=True, bijection=False)
    many.print(opt)
    print_multiline_escape(many.repr(opt))
    expected = """
0 Add
  1 Add
    2 [] Scalar 1 # \x1b[35m0\x1b[0m
    2 [] Scalar 1 # \x1b[35m1\x1b[0m
  2 [] Scalar 1 # \x1b[35m2\x1b[0m
  1 Add
    2 [] Scalar 1 # \x1b[35m3\x1b[0m
    2 [] Scalar 1 # \x1b[35m4\x1b[0m""".strip(
        "\n"
    )

    assert many.repr(opt) == expected
    rc.nest_adds(many, (((3, 2), 0), (1, 4))).print()  # nesting works with these numbers

    trav = rc.new_traversal(end_depth=2)
    opt = opt.evolve(traversal=trav)
    many.print(opt)
    print_multiline_escape(many.repr(opt))
    expected = """
0 Add
  1 Add ... # \x1b[35m0\x1b[0m
  2 [] Scalar 1 # \x1b[35m1\x1b[0m
  1 Add ... # \x1b[35m2\x1b[0m""".strip(
        "\n"
    )
    assert many.repr(opt) == expected
    rc.nest_adds(many, (1, 0, 2), traversal=trav).print()  # nesting works with these numbers


def test_parse_named_axes():
    printer = rc.PrintOptions(show_all_named_axes=True)
    einsum = """0 [large:3] Einsum ab,ac->a
  1 [large:3,small:1] Scalar 1
  2 [large:3,small:1] Scalar 1.5"""
    assert printer.repr(rc.Parser()(einsum)) == einsum
    einsum = """0 [small:1,large:3] Einsum ab,cb->bc
  1 [large:3,small:1] Scalar 1
  2 [large:3,small:1] Scalar 1.5"""
    assert printer.repr(rc.Parser()(einsum)) == einsum
    einsum = """0 [small:1,large:3] Einsum ab,cb->bc
  1 [3,1] Scalar 1
  2 [large:3,small:1] Scalar 1.5"""
    assert printer.repr(rc.Parser()(einsum)) == einsum
    einsum = """0 [small:1,3] Einsum ab,cb->bc
  1 [3,1] Scalar 1
  2 [3,small:1] Scalar 1.5"""
    assert printer.repr(rc.Parser()(einsum)) == einsum
    einsum = """0 [small:1,large:3] Einsum fancy: largerrr small, large small -> small large
  1 [largerrr:3,small:1] Scalar 1
  2 [large:3,small:1] Scalar 1.5"""
    assert printer.repr(rc.Parser()(einsum)) == einsum
    einsum = """0 [small:1,large:3] Einsum fancy: largerrr small, large small -> small large
  1 [largerrr:3,small:1] Scalar 1
  2 [large:3,smaller:1] Scalar 1.5"""
    assert printer.repr(rc.Parser()(einsum)) == einsum


def test_different_color_for_common_nodes():
    f = rc.PrintHtmlOptions.type_colorer()

    colors = [
        f(rc.Add()),
        f(rc.Concat(rc.Scalar(1.0, (5,)), axis=0)),
        f(rc.Einsum.scalar_mul(rc.Add(), 2)),
        f(rc.Array(torch.zeros(5))),
        f(rc.sigmoid(rc.Array(torch.zeros(5)))),
        f(rc.Rearrange.from_string(rc.Array(torch.zeros(5)), "a -> a")),
    ]
    assert len(colors) == len(set(colors))


def test_print_arrows_and_only_child_below():  # regression test for #1699
    printer = rc.PrintOptions(arrows=True, only_child_below=True)

    circ = """0 Add\n  1 [1] Scalar 0.7"""
    printer.print(rc.Parser()(circ))
    assert printer.repr(rc.Parser()(circ)) == """0 Add\n▼\n1 [1] Scalar 0.7"""

    circ = """0 Add
  1 Add
    2 [1] Scalar 0.7
  3 Add
    4 [1] Scalar 0.5
    5 [1] Scalar 6.3"""
    printer.print(rc.Parser()(circ))
    assert (
        printer.repr(rc.Parser()(circ))
        == """0 Add
├‣1 Add
│ ▼
│ 2 [1] Scalar 0.7
└‣3 Add
  ├‣4 [1] Scalar 0.5
  └‣5 [1] Scalar 6.3"""
    )

    circ = """0 Add
  1 Add
    2 [1] Scalar 0.5
    3 [1] Scalar 6.3
  4 Add
    5 [1] Scalar 0.7"""
    printer.print(rc.Parser()(circ))
    assert (
        printer.repr(rc.Parser()(circ))
        == """0 Add
├‣1 Add
│ ├‣2 [1] Scalar 0.5
│ └‣3 [1] Scalar 6.3
└‣4 Add
  ▼
  5 [1] Scalar 0.7"""
    )


def test_parse_cycle_error():
    with pytest.raises(rc.ParseCircuitCycleError):
        P(
            """
1 Add
  2 Add
    1
        """
        )



import pytest
import torch

import rust_circuit as rc
import rust_circuit.module_library as mod_l
from rust_circuit import (
    OptimizationSettings,
    TorchDeviceDtypeOp,
    batch_to_concat,
    cast_circuit,
    optimize_and_evaluate,
    substitute_all_modules,
)
from rust_circuit.py_utils import timed


@pytest.mark.cuda
@pytest.mark.parametrize(
    "config",
    [
        {"num_layers": 6, "head_size": 128, "num_heads": 4},
        {"num_layers": 12, "head_size": 64, "num_heads": 12},
        {"num_layers": 16, "head_size": 64, "num_heads": 16},
    ],
)
@pytest.mark.skip("model generation too slow")
def test_big_models_cuda(config):
    raw_test_big_models(config, device="cuda")


@pytest.mark.parametrize("config", [{"num_layers": 4, "head_size": 5, "num_heads": 3}])
def test_big_models(config):
    raw_test_big_models(config, device="cpu")


@pytest.mark.cuda
def test_big_wide():
    raw_test_big_models({"num_layers": 6, "head_size": 128, "num_heads": 8}, device="cuda", batch_size=200)


@pytest.mark.cuda
def test_tall_tiny_batch():
    raw_test_big_models(
        {"num_layers": 12, "head_size": 2, "num_heads": 2},
        device="cuda",
        batch_size=64,
        concat_batch_size=64,
        seq_len=2,
    )


@pytest.mark.cuda
def test_big_batch():
    raw_test_big_models(
        {"num_layers": 4, "head_size": 128, "num_heads": 8}, device="cuda", batch_size=32, concat_batch_size=32
    )


def raw_test_big_models(config, device="cpu", batch_size=1_000, concat_batch_size=None, seq_len=128):
    with timed("model creation"):
        y_rust, _, _ = mod_l.TransformerParams(
            mod_l.TransformerBlockParams(mlp_act_type="gelu", attn_bias=True, mlp_output_bias=True),
            num_layers=config["num_layers"],
        ).garbage_call(
            **{s: c for s, c in config.items() if s != "num_layers"},
            seq_len=seq_len,
            batch_shape=(batch_size,),
        )
        y_rust = substitute_all_modules(y_rust)

    cast_circuit(y_rust, TorchDeviceDtypeOp(device, "float32"))
    with timed("eval"):
        if concat_batch_size is not None:
            with timed("batching"):
                y_rust = batch_to_concat(y_rust, 0, concat_batch_size)
        optimize_and_evaluate(y_rust, OptimizationSettings(verbose=3, max_memory=2_000_000_000, scheduling_naive=False))


@pytest.mark.cuda
def test_len4():
    with open("/home/ubuntu/rrfs/adria/gpt2small/seq_len_4.txt") as f:
        circuit_raw = rc.cast_circuit(rc.Parser()(f.read()), rc.TorchDeviceDtypeOp(device="cpu"))
    # circuit_raw.print()
    circuit_arr = rc.Updater(
        lambda _: rc.DiscreteVar.new_uniform(rc.Array(torch.randint(0, 1, (100, 4), dtype=torch.float32)))
    )(circuit_raw, "tokens")
    # circuit_arr.print()
    circuit_sampled = rc.Sampler(rc.RandomSampleSpec()).sample(circuit_arr)
    # warm up
    circuit_sampled.evaluate()
    circuit_sampled.evaluate()
    with timed(".evaluate"):
        circuit_sampled.evaluate()
    # circuit_sampled.print()
    print("now opevaling")
    print(rc.count_nodes(circuit_sampled))
    with timed("opteval"):
        optimize_and_evaluate(circuit_sampled, OptimizationSettings(verbose=2))


if __name__ == "__main__":
    # raw_test_big_models({"num_layers": 50, "head_size": 16, "num_heads": 4}, "cuda")
    test_len4()



from datetime import timedelta
from uuid import UUID

import hypothesis
import hypothesis.extra.numpy as st_np
import pytest
import torch
from torch import tensor

from interp.circuit.cum_algo import eps_attrib
from interp.circuit.testing.topdown_circuit import CircuitProbs as CP
from rust_circuit import Add, Array
from rust_circuit import Circuit as rCircuit
from rust_circuit import Concat
from rust_circuit import Cumulant
from rust_circuit import Cumulant as rCumulant
from rust_circuit import DiscreteVar
from rust_circuit import DiscreteVar as rDiscreteVar
from rust_circuit import Einsum
from rust_circuit import Einsum as rEinsum
from rust_circuit import GeneralFunction
from rust_circuit import GeneralFunction as rGeneralFunction
from rust_circuit import Index, Rearrange, RunDiscreteVarAllSpec, Sampler, Scalar
from rust_circuit import StoredCumulantVar as rStoredCumulantVar
from rust_circuit import Tag, rewrite_cum_to_circuit_of_cum
from rust_circuit.cum_algo import eps_attrib as r_eps_attrib
from rust_circuit.interop_rust import py_to_rust, rust_to_py

from .util import make_discrete_var, st_Circuit


@hypothesis.example(  # Two DiscreteVars
    cumulant=(
        lambda c0: Cumulant(
            DiscreteVar(
                values=Scalar(0.0, shape=(1, 1), name=""),
                probs_and_group=Tag(
                    node=c0, uuid=UUID("51b47e37-4766-41dd-9f2d-9f0a8c3742a4"), name="uniform_probs_and_group"
                ),
                name="",
            ),
            DiscreteVar(
                values=Scalar(0.0, shape=(1, 2), name=""),
                probs_and_group=Tag(
                    node=c0, uuid=UUID("75ebd054-4fd9-4086-b215-2ac363445cf8"), name="uniform_probs_and_group"
                ),
                name="",
            ),
            name="",
        )
    )(Scalar(value=1.0, shape=(1,), name="uniform_probs")),
)
@hypothesis.example(  # 4 DiscreteVars
    cumulant=(
        lambda c0: Cumulant(
            *(
                DiscreteVar(
                    values=Scalar(i / 4, shape=(1, i + 4), name=""),
                    probs_and_group=c0,
                )
                for i in range(1, 5)
            ),
            name="",
        )
    )(Scalar(value=1.0, shape=(1,), name="uniform_probs")),
)
@hypothesis.example(
    cumulant=Cumulant(
        Index(
            node=Einsum(
                (Scalar(1.0, shape=(2, 1), name=""), (0, 1)),
                (
                    Array(
                        value=tensor([1.5410]),
                        name="",
                    ),
                    (2,),
                ),
                out_axes=(0, 1),
                name="",
            ),
            index=(0, slice(0, 1, None)),
            name="",
        ),
        DiscreteVar(
            values=Scalar(1.0, shape=(2,), name=""),
            probs_and_group=Tag(
                node=Scalar(value=0.5, shape=(2,), name="uniform_probs"),
                uuid=UUID("de5bfc0e-b11d-4c42-93bf-315db41f587a"),
                name="uniform_probs_and_group",
            ),
            name="",
        ),
        name="",
    ),
)
@hypothesis.example(
    cumulant=(
        lambda c0: Cumulant(
            c0,
            DiscreteVar(
                values=Scalar(0.0, shape=(2, 1), name=""),
                probs_and_group=Tag(
                    node=Scalar(value=0.5, shape=(2,), name="uniform_probs"),
                    uuid=UUID("79afd6c2-c9a8-4f5b-a258-95ba1bc6a10f"),
                    name="uniform_probs_and_group",
                ),
                name="",
            ),
            c0,
            Scalar(1.0, shape=(4,), name=""),
            Scalar(0.0, shape=(1, 1), name=""),
            Scalar(1.0, shape=(2,), name=""),
            Scalar(0.0, shape=(5,), name=""),
            name="",
        )
    )(Scalar(0.0, shape=(), name="A")),
)
@hypothesis.example(  # Tag
    cumulant=Cumulant(
        Tag(
            node=DiscreteVar(
                values=Scalar(0.0, shape=(1,), name=""),
                probs_and_group=Tag(
                    node=Scalar(value=1.0, shape=(1,), name="uniform_probs"),
                    uuid=UUID("cf4f2809-1c44-4947-ba57-f7bf6b4d9a26"),
                    name="uniform_probs_and_group",
                ),
                name="",
            ),
            uuid=UUID("e3e70682-c209-4cac-629f-6fbed82c07cd"),
            name="",
        ),
        name="",
    ),
)
@hypothesis.example(
    cumulant=Cumulant(
        Einsum(
            (
                DiscreteVar(
                    values=Scalar(1.0, shape=(1, 2), name=""),
                    probs_and_group=GeneralFunction.new_by_name(
                        Scalar(0.0, shape=(1,), name=""),
                        spec_name="softmax",
                        name="_softmax",
                    ),
                    name="",
                ),
                (0,),
            ),
            out_axes=(),
            name="",
        ),
        name="",
    ),
)
@hypothesis.example(
    cumulant=Cumulant(
        Einsum((Scalar(1.0, shape=(1,), name=""), (0,)), out_axes=(0, 0), name=""),
        name="",
    ),
)
@hypothesis.example(
    cumulant=Cumulant(
        Rearrange.from_string(
            Einsum(
                (
                    Rearrange.from_string(
                        Concat(
                            Scalar(1.0, shape=(2, 1, 1, 1, 1), name=""),
                            Index(
                                node=DiscreteVar(
                                    values=Scalar(0.0, shape=(3, 1, 1, 1, 1, 1), name=""),
                                    probs_and_group=Tag(
                                        node=Scalar(value=0.3333333333333333, shape=(3,), name="uniform_probs"),
                                        uuid=UUID("6ab46340-445c-4633-8aff-e6634e336d57"),
                                        name="uniform_probs_and_group",
                                    ),
                                    name="",
                                ),
                                index=(
                                    slice(0, 1, None),
                                    slice(0, 1, None),
                                    slice(0, 1, None),
                                    slice(0, 1, None),
                                    slice(0, 1, None),
                                ),
                                name="",
                            ),
                            axis=0,
                            name="",
                        ),
                        "s0 s1 s2 s3 s4 -> s0 s1 s2 s3 s4",
                    ),
                    (0, 1, 2, 3, 4),
                ),
                out_axes=(),
                name="",
            ),
            "->",
        ),
        Scalar(1.0, shape=(1,), name=""),
        name="",
    ),
)
@hypothesis.example(
    cumulant=Cumulant(
        Add.from_weighted_nodes(
            (
                Index(
                    node=Scalar(1.0, shape=(), name=""),
                    index=(),
                    name="",
                ),
                1.0,
            ),
            (
                DiscreteVar(
                    values=Scalar(1.0, shape=(1,), name=""),
                    probs_and_group=Tag(
                        node=Scalar(value=1.0, shape=(1,), name="uniform_probs"),
                        uuid=UUID("b95463a1-b440-4d3e-a7cf-5732b90a680a"),
                        name="uniform_probs_and_group",
                    ),
                    name="",
                ),
                0.0,
            ),
        ),
        name="",
    ),
)
@hypothesis.given(
    cumulant=st_Circuit(
        st_np.array_shapes(min_dims=0, max_dims=3),
        max_growth_steps=10,
        probs_per_depth=[
            CP.kw(all=0, Cumulant=1),
            # Only allow DiscreteVars as random variables, otherwise draw some branches. Do not draw GeneralFunctions.
            CP.kw(all=1, StoredCumulantVar=0, DiscreteVar=5, leaves=1, branches=5, randoms=0, GeneralFunction=0),
        ],
        # At depth >2, GeneralFunctions are now fair game
        probs_default=CP.kw(all=1, StoredCumulantVar=0, DiscreteVar=5, leaves=1, branches=5, randoms=0),
    )
)
@hypothesis.settings(deadline=timedelta(seconds=3))
@pytest.mark.skip(reason="TODO unskip; Bug found by fuzzing")
def test_rewrite_cum_to_circuit_of_cum(cumulant: Cumulant):
    raw_test_rewrite_cum_to_circuit_of_cum(cumulant)


def raw_test_rewrite_cum_to_circuit_of_cum(cumulant: rCumulant):
    to_expand = cumulant.children[0]

    hypothesis.assume(
        not isinstance(to_expand, (rGeneralFunction, rDiscreteVar, rStoredCumulantVar)),
        "Do not try to expand things that cannot be expanded",
    )

    hypothesis.assume(
        not isinstance(to_expand, rEinsum) or len(to_expand.args) <= 3,
        "Expanding an Einsum with more than 3 children is too expensive",
    )

    any_constant_zero = any(c.is_constant for c in cumulant.children) and cumulant.num_children > 1
    hypothesis.assume(not any_constant_zero, "trivial if more than 1 inputs are constant")

    hypothesis.assume(
        cumulant.num_children <= 3 or to_expand.is_constant, "only cumulants of order <= 3 -- otherwise too expensive"
    )

    substituted_cumulant: rCircuit = rewrite_cum_to_circuit_of_cum(cumulant, to_expand, lambda _: "")
    assert not (isinstance(substituted_cumulant, rCumulant) and not isinstance(to_expand, rCumulant))

    sample_spec = RunDiscreteVarAllSpec.create_full_from_circuits(cumulant)
    try:
        sampled_cumulant = Sampler(sample_spec)(cumulant)
    except:
        hypothesis.reject()  # circuit too weird to be sampled
        # For example
        # 0 '' [] Cumulant
        #   1 '' [] Rearrange  ->
        #     2 '' [] DiscreteVar
        #     3 '' [1] Cumulant
        #         4 '' [1] Scalar 0
        #     5 '_softmax' [1] GeneralFunction softmax
        #         3

    sampled_substituted_cumulant = Sampler(sample_spec)(substituted_cumulant)

    atol = 1e-6
    rtol = 1e-5
    torch.testing.assert_allclose(
        sampled_cumulant.evaluate(), sampled_substituted_cumulant.evaluate(), atol=atol, rtol=rtol
    )


def compare_eps_attrib_values(c: Cumulant):
    for non_centered in [False, True]:
        to_expand = py_to_rust(eps_attrib(rust_to_py(c), non_centered=non_centered))
        r_to_expand = r_eps_attrib(c, non_centered=non_centered)

        sample_spec = RunDiscreteVarAllSpec.create_full_from_circuits(to_expand)
        sampled = Sampler(sample_spec)(to_expand).evaluate()
        r_sample_spec = RunDiscreteVarAllSpec.create_full_from_circuits(r_to_expand)
        # TODO: make sampling work?
        r_sampled = Sampler(r_sample_spec)(r_to_expand.cast_module().substitute()).evaluate()

        atol = 1e-6
        rtol = 1e-5
        torch.testing.assert_allclose(sampled, r_sampled, atol=atol, rtol=rtol)


def test_eps_attrib_values():
    x, _, _ = make_discrete_var(values_shape=(2, 3), name="x", batch_size=17)
    y, _, _ = make_discrete_var(values_shape=(3, 2), name="y", batch_size=17, group=x.probs_and_group)
    z, _, _ = make_discrete_var(values_shape=(4,), name="z", batch_size=17, group=x.probs_and_group)

    vals = [
        Cumulant(),
        Cumulant(x),
        Cumulant(x, z),
        Cumulant(z, x),
        Cumulant(y, z),
        Cumulant(x, y),
        Cumulant(x, y, z),
    ]

    for val in vals:
        compare_eps_attrib_values(val)


if __name__ == "__main__":
    test_eps_attrib_values()
    # import rust_circuit as rc
    # s, *_ = rc.symbolic_sizes()
    # rc.Symbol.new_with_random_uuid((s,)).print()



import rust_circuit as rc


def test_set_of_identities():
    setty = rc.SetOfCircuitIdentities()
    circ = rc.Array.randn(10)
    setty.insert(circ)
    print(len(setty))
    print(circ in setty)
    print(rc.Array.randn(11) in setty)
    setty.insert(circ)
    print(len(setty))



import rust_circuit.optional as op

from . import _rust as rc


def get_matching_nodes_and_args(mod: rc.Module, sym_matcher: rc.MatcherIn):
    sym_matcher = rc.Matcher(sym_matcher)
    nodes_arg_specs = [
        (node, arg_spec) for node, arg_spec in zip(mod.nodes, mod.spec.arg_specs) if sym_matcher(arg_spec.symbol)
    ]
    nodes = [n for n, _ in nodes_arg_specs]
    arg_specs = [a for _, a in nodes_arg_specs]
    return nodes, arg_specs


def filter_module_inputs(mod: rc.Module, sym_matcher: rc.MatcherIn):
    nodes, arg_specs = get_matching_nodes_and_args(mod, sym_matcher)
    return rc.Module.new_flat(
        rc.ModuleSpec(mod.spec.circuit, arg_specs, False, False),
        *nodes,
        name=op.map(mod.op_name, lambda x: x + " drop")
    )


def drop_module_inputs(mod: rc.Module, drop_sym_matcher: rc.MatcherIn):
    """filter_module_inputs(mod, ~rc.Matcher(drop_sym_matcher))"""
    return filter_module_inputs(mod, ~rc.Matcher(drop_sym_matcher))



import pytest
import torch
from torch.testing import assert_close

from rust_circuit import FINISHED, Add, Array, Einsum, IterateMatchResults, IterativeMatcher, Matcher
from rust_circuit import NestAddsSpecInfo as AddInfo
from rust_circuit import NestEinsumsSpec
from rust_circuit import NestEinsumsSpecInfo as EinInfo
from rust_circuit import (
    NestMatcher,
    NestMatcherMatchedMultipleAndMustBeUniqueError,
    NestPermutationMissesIdxsAndNoRestInSpecError,
)
from rust_circuit import NestRest as Rest
from rust_circuit import (
    PermIntsNotUniqueError,
    PrintOptions,
    einsum_flatten,
    nest_adds,
    nest_einsums,
    new_traversal,
    restrict,
)

# os.environ["RUST_BACKTRACE"] = "1"


# %% [markdown]

# # Basic idea

# In this demo, we'll look at nesting rewrites.
# These are powerful and convenient rewrites for circuit types
# which can be manipulated into trees. Specifically, we can nest einsums and adds.
#
# We'll be looking at nest_einsums and nest_adds

# (This is the same as the original rearrange_muls rewrite in python, but with a decent number of additional features)

# %%

# first, we'll create a flat einsum we want to rearrange
a, b, c, d, f, e = 2, 3, 4, 5, 6, 7
x = Einsum.from_einsum_string(
    "a b c, b c d, a e, b d, f c, a c d, e a f -> a c d f",
    *(
        Array(torch.randn(*shape, dtype=torch.float64), name=name)
        for (shape, name) in [
            ((a, b, c), "abc"),
            ((b, c, d), "bcd"),
            ((a, e), "ae"),
            ((b, d), "bd"),
            ((f, c), "ec"),
            ((a, c, d), "acd"),
            ((e, a, f), "eaf"),
        ]
    )
).normalize_ints()

x.print()

# %%

# nest einsums take a nested sequence structure.
# Each integer refers to an index into the flattened einsum.

# So here, we group the first 3 arguments in one subset and the last 4 arguments into another subset
z = nest_einsums(x, ((0, 1, 2), (3, 4, 5, 6)))
assert_close(x.evaluate(), z.evaluate())
z.print()

# %%


# Here's a more complex setup
z = nest_einsums(x, ((5, (1, 3)), 2, ((4,), 0), 6))
assert_close(x.evaluate(), z.evaluate())
z.print()

# %%

# ints must be a permutation of all leaf locations
with pytest.raises(NestPermutationMissesIdxsAndNoRestInSpecError):
    nest_einsums(x, (0, 2))
with pytest.raises(PermIntsNotUniqueError):
    nest_einsums(x, (0, 1, 1, 2, 3, 4, 5, 6))
with pytest.raises(PermIntsNotUniqueError):
    nest_einsums(x, (0, 1, 2, 3, 4, (5, 2), 6))

# %%

# this also works for adds!

x_add = Add(
    *(
        Array(torch.randn(*shape, dtype=torch.float64), name=name)
        for (shape, name) in [
            ((a, b, c), "abc"),
            ((a, 1, 1), "a11"),
            ((1, b, 1), "1b1"),
            ((b, c), "bc"),
            ((c,), "c"),
            ((1, c), "1c"),
        ]
    )
)
x_add.print()

z_add = nest_adds(x_add, ((0, 4, 2), (3, 1, 5)))
assert_close(x_add.evaluate(), z_add.evaluate())
z_add.print()

# we'll proceed using einsums, but all of this applies to adds also!

# %%

# you can print with the corresponding numbers
my_printer = PrintOptions(number_leaves=True, traversal=new_traversal(term_early_at=~Matcher(Einsum)), bijection=False)
# the 'number_leaves' print option implies printing as a tree (not a dag), so
# be carefully, this might be very big if you don't term_early_at
my_printer.print(x)

# you can also use
PrintOptions().einsum_nest_default().print(x)

# there's also
PrintOptions().add_nest_default().print(x_add)

# going forward we'll use this printer
opt = PrintOptions().einsum_nest_default()


# %% [markdown]

# # Additional Feats

# That's pretty much the entire basic idea, but there are a bunch of additional features which can be quite useful.

# %% [markdown]

# ## 'Rest'

# %%


# You can specify a 'rest' argument which groups all arguments not otherwise handled.
z = nest_einsums(x, ((5, (1, 3)), 2, Rest()))
assert_close(x.evaluate(), z.evaluate())
opt.print(z)

# %%

# you can also make this 'flatten' one level inline (imagine this as 'splatting' the rest (splat is the '*' python operator))
z_rest = nest_einsums(x, ((5, (1, 3)), 2, Rest(flat=True)))
assert_close(x.evaluate(), z_rest.evaluate())
opt.print(z_rest)

# %%

# If Rest matches nothing, it will be flattened in
z = nest_einsums(x, (0, 1, 2, 3, 4, 5, 6, Rest()))
assert_close(x.evaluate(), z.evaluate())
opt.print(z)


# %% [markdown]

# ## Matching instead of numbering

# %%

# numbering can some times be annoying, so we can instead pass a matcher
z_new = nest_einsums(x, (("acd", ("bcd", Matcher.regex("^bd$"))), 2, Rest(flat=True)))
# (this is any IterativeMatcherIn, the iterative matcher is run starting at the root einsum)
# warning: in python, 'Regex' matched starts, but in rust it can match anywhere. So we anchor with '^'!
assert z_new == z_rest  # exactly the same as previous example

# %%

# this matcher has to match exactly one thing!!!
with pytest.raises(NestMatcherMatchedMultipleAndMustBeUniqueError):
    nest_einsums(x, (("acd", Matcher.regex("b")), 2, Rest(flat=True)))

# %% [markdown]

# ## Matching multiple with a matcher

# %%

# we can also explicitly pass in a matcher which can match multiple via the `NestMatcher` class.
# this has roughly the same semantics as `Rest` including the ability to specify `flat=True`
z = nest_einsums(x, (NestMatcher(Matcher.regex("b")), Rest(flat=True)))
assert_close(x.evaluate(), z.evaluate())
opt.print(z)
# (we could change things so that a matcher which isn't wrapped (like previous
# section) can match multiple, but this doesn't seem like right default IMO)

# %%

z = nest_einsums(x, ((NestMatcher(Matcher.regex("b"), flat=True), 2), Rest(flat=True)))
assert_close(x.evaluate(), z.evaluate())
opt.print(z)

# %%

# there are some validation args also
z = nest_einsums(
    x,
    (
        (
            NestMatcher({"bd", "ae"}, fancy_validate=True),  # checks that all things in the set indeed match something
            NestMatcher("abc", assert_unique=True),
            NestMatcher("347387", assert_exists=False),  # matches nothing, flattened in
        ),
        Rest(flat=True),
    ),
)
assert_close(x.evaluate(), z.evaluate())
opt.print(z)


# %% [markdown]

# ## Partial traversals

# %%

# sometimes we have deep einsum trees and we only want to traverse down part of
# the way and rearrange up to the that point.

# for example:
opt.print(z_rest)
# here, let's say we don't want to rearrange bcd * bd


# a traversal is just an iterative matcher where we *only* care about whether or not it has terminated.
print_traversal = restrict(opt.traversal, term_early_at="bcd * bd")

# note that the numbers here are the ones obtained when partially traversing! you might want to use matchers
# here, or print with the same filter as you are traversing with.
opt.evolve(traversal=print_traversal).print(z_rest)

z_filter = nest_einsums(z_rest, (1, 2, 3, 0, 4, 5), traversal=print_traversal)
assert_close(x.evaluate(), z_filter.evaluate())
opt.evolve(traversal=print_traversal).print(z_filter)

# %%

# the previous example can also be done via setting an end_depth
depth_traverse = restrict(opt.traversal, end_depth=3)

opt.evolve(traversal=depth_traverse).print(z_rest)
z_filter_new = nest_einsums(z_rest, (1, 2, 3, 0, 4, 5), traversal=depth_traverse)
assert z_filter_new == z_filter

# %% [markdown]

# ## Setting node info

# it can often be nice to set names etc on a given group.
# There's a nice API for this!

# %%

# we can specify names inside the spec with EinInfo
z_named = nest_einsums(x, ((5, EinInfo((1, 3), name="my_new_name")), 2, EinInfo(Rest(), name="named_rest")))
assert_close(x.evaluate(), z_named.evaluate())
_ = Matcher("my_new_name").get_unique(z_named)
_ = Matcher("named_rest").get_unique(z_named)
opt.print(z_named)

# %%

# We can also permute the output axes of sub einsums
z_perm = nest_einsums(
    x, ((5, EinInfo((1, 3), name="my_new_name", out_axes_perm=[2, 0, 1])), 2, EinInfo(Rest(), name="named_rest"))
)
assert_close(x.evaluate(), z_perm.evaluate())
assert_close(
    Matcher("my_new_name").get_unique(z_perm).evaluate(),
    Matcher("my_new_name").get_unique(z_named).rearrange_str("a b c -> c a b").evaluate(),
)
_ = Matcher("named_rest").get_unique(z_perm)
opt.print(z_perm)

# %%

# nest_adds just support setting the name (because they don't have output axes to permute)
z_add = nest_adds(x_add, ((5, AddInfo((1, 3), name="my_new_name")), 2, AddInfo(Rest(), name="named_rest")))
assert_close(x_add.evaluate(), z_add.evaluate())
_ = Matcher("my_new_name").get_unique(z_add)
_ = Matcher("named_rest").get_unique(z_add)
opt.print(z_add)


# %% [markdown]

# ## Retaining node names

# generally when doing rewrites, it's useful to keep names and other node properties when possible.
# nest_{einsums/adds} does this!

# %%

# let's first make an einsum with a bunch of names
z_with_names = nest_einsums(x, (4, EinInfo((3, EinInfo((2, 1), "other"), EinInfo(((0, 5), 6), "here")), "outer")))
assert_close(x.evaluate(), z_with_names.evaluate())
print()
opt.print(z_with_names)

z_here, z_other, z_outer = [Matcher(n).get_unique(z_with_names) for n in ["here", "other", "outer"]]

# %%


# if we nest things such that some subnode could be made extensionally equal (if its children are the same numbers),
# the name and output axis ordering will be retained!
z_keep_here = nest_einsums(z_with_names, (0, (1, 2, 3, (6, 4, 5))))
assert_close(x.evaluate(), z_keep_here.evaluate())
print()
opt.print(z_keep_here)
assert_close(Matcher("here").get_unique(z_keep_here).evaluate(), z_here.evaluate())
assert_close(Matcher("outer").get_unique(z_keep_here).evaluate(), z_outer.evaluate())
assert Matcher("other").get_unique_op(z_keep_here) is None

# %%

# We can also do this with rest and matchers (this example is silly, but you get the idea
z_keep_via_match = nest_einsums(
    z_with_names, (0, (NestMatcher({"abc", "acd", "eaf"}, flat=True), NestMatcher({"ae", "bcd"}, flat=True), 1))
)
assert_close(x.evaluate(), z_keep_via_match.evaluate())
print()
opt.print(z_keep_via_match)
assert_close(Matcher("here").get_unique(z_keep_via_match).evaluate(), z_here.evaluate())
assert_close(Matcher("outer").get_unique(z_keep_via_match).evaluate(), z_outer.evaluate())
assert_close(Matcher("other").get_unique(z_keep_via_match).evaluate(), z_other.evaluate())

# %%

z_keep_via_rest = nest_einsums(z_with_names, (0, (Rest(flat=True), NestMatcher({"ae", "bcd"}, flat=True), 1)))
assert z_keep_via_rest == z_keep_via_match

# %% [markdown]

# ## Shrinking nodes with extra retained dims (not very important)

# %%


# this retaining will (by default) retain extra axes which could be reduced
sub_ein = Einsum.from_einsum_string(
    "eab,ae->ab", Array.randn(e, a, b, name="eab"), Array.randn(a, e, name="ae"), name="sub"
)
with_sub_ein = Einsum.from_einsum_string(
    "ab,ac->c",
    sub_ein,
    Array.randn(a, c, name="ac"),
)
# here, we could sum out the 'b' dim in the above eab,ae->ab, but we retain it for equivalence with original node
unneeded_retain = nest_einsums(
    with_sub_ein,
    (2, (1, 0)),
)
assert_close(Matcher("sub").get_unique(unneeded_retain).evaluate(), sub_ein.evaluate())
opt.print(unneeded_retain)

# %%

# but if you'd like you can manually force these dims to be 'shrunk'
unneeded_retain_shrink = nest_einsums(
    with_sub_ein,
    (
        2,
        EinInfo((1, 0), shrink_out_axes=True, name="sub_new"),
    ),  # note that we had to specify the name again. The name would have been autorenamed if we didn't set this.
)
assert_close(Matcher("sub_new").get_unique(unneeded_retain_shrink).evaluate(), sub_ein.evaluate().sum(dim=1))
opt.print(unneeded_retain_shrink)

# %% [markdown]

# ## More tests

# From here on out, there are some more tests/demos. Feel free to ignore!


# %%


# now we can verify various changes have no effect
y = einsum_flatten(x)  # flattening has no effect
assert x == y
y = nest_einsums(x, (0, 1, 2, 3, 4, 5, 6)).normalize_ints()  # maybe this should normalize ints by default?
assert x == y

# %%


def print_multiline_escape(s: str):
    uuid = "6c6805a0-3aaf-492c-93a3-00851eacaacf"
    print(s.replace("\n", uuid).encode("unicode_escape").decode("ASCII").replace(uuid, "\n"))


z = nest_einsums(x, ((5, (1, 3)), 2, ((4,), 0), 6))
assert_close(x.evaluate(), z.evaluate())
print()
opt.print(z)
print()
print_multiline_escape(opt.repr(z))

s = """
abc * bcd * ae * bd * ec * acd * eaf Einsum acef,ad,abce,dab->acfb
  acd * bcd * bd Einsum acf,cef->acef
    acd [2,4,5] Array # \x1b[35m0\x1b[0m
    bcd * bd Einsum ecf,ef->cef
      bcd [3,4,5] Array # \x1b[35m1\x1b[0m
      bd [3,5] Array # \x1b[35m2\x1b[0m
  ae [2,7] Array # \x1b[35m3\x1b[0m
  ec * abc Einsum bc,aec->abce
    0 ec Einsum bc->bc
      1 ec [6,4] Array # \x1b[35m4\x1b[0m
    abc [2,3,4] Array # \x1b[35m5\x1b[0m
  eaf [7,2,6] Array # \x1b[35m6\x1b[0m""".strip(
    "\n"
)
assert opt.repr(z) == s


# %%

z = nest_einsums(x, (4, Rest()))
assert_close(x.evaluate(), z.evaluate())
print()
z.print()


# %%

z = nest_einsums(x, (4, (1, 2, Rest())))
assert_close(x.evaluate(), z.evaluate())
print()
z.print()


# %%

z = nest_einsums(x, Rest())
assert_close(x.evaluate(), z.evaluate())
print()
z.print()

# %%

z = nest_einsums(x, Rest(flat=True))
assert_close(x.evaluate(), z.evaluate())
print()
z.print()


# %%

z_no_flat = nest_einsums(x, (4, (1, 2, Rest(flat=True))))
assert_close(x.evaluate(), z_no_flat.evaluate())
print()
z_no_flat.print()

# %%

z = nest_einsums(x, (4, (1, 2, Rest(flat=True))))
assert_close(x.evaluate(), z.evaluate())
print()
z.print()

# %%

z_prev = nest_einsums(x, (4, (3, 2, 1, ((0, 5), 6))))
assert_close(x.evaluate(), z_prev.evaluate())
print()
z_prev.print()

# %%

z = nest_einsums(x, (4, (3, 2, 1, (EinInfo((0, 5), out_axes_perm=(2, 3, 1, 0)), 6))))
assert_close(x.evaluate(), z.evaluate())
print()
z.print()

# %%

sub_new = Einsum.from_einsum_string(
    "a b c, b c d, a e -> e",
    *(
        Array(torch.randn(*shape, dtype=torch.float64), name=name)
        for (shape, name) in [
            ((a, b, c), "abc"),
            ((b, c, d), "bcd"),
            ((a, e), "ae"),
        ]
    )
).normalize_ints()
sub_new.print()
sub = nest_einsums(sub_new, (0, (1, 2))).normalize_ints()
sub_fin = nest_einsums(sub, ((1, 2), 0))

# %%


# %%

z_with_names = nest_einsums(
    x, (4, EinInfo((3, EinInfo((2, 1), "other"), EinInfo(((0, 5), 6), "here")), "outer"))
).normalize_ints()
assert_close(x.evaluate(), z_with_names.evaluate())
print()
z_with_names.print()

z_here, z_other, z_outer = [Matcher(n).get_unique(z_with_names) for n in ["here", "other", "outer"]]

# %%


z_with_names.print()
einsum_flatten(z_with_names).all_input_axes()
z_keep_here = nest_einsums(z_with_names, (0, (1, 2, 3, (6, 4, 5))))
assert_close(x.evaluate(), z_keep_here.evaluate())
print()
z_keep_here.print()
assert_close(Matcher("here").get_unique(z_keep_here).evaluate(), z_here.evaluate())
assert_close(Matcher("outer").get_unique(z_keep_here).evaluate(), z_outer.evaluate())
assert Matcher("other").get_unique_op(z_keep_here) is None

# %%

z_keep_other = nest_einsums(z_with_names, (0, ((2, 3), (6, 4, 1, 5))))
assert_close(x.evaluate(), z_keep_other.evaluate())
print()
z_keep_other.print()
assert_close(Matcher("other").get_unique(z_keep_other).evaluate(), z_other.evaluate())
assert_close(Matcher("outer").get_unique(z_keep_other).evaluate(), z_outer.evaluate())
assert Matcher("here").get_unique_op(z_keep_other) is None

# %%

z_keep_just_here = nest_einsums(z_with_names, (0, (2, 1, 3), (6, 4, 5)))
assert_close(x.evaluate(), z_keep_just_here.evaluate())
print()
z_keep_just_here.print()
assert_close(Matcher("here").get_unique(z_keep_here).evaluate(), z_here.evaluate())
assert Matcher("other").get_unique_op(z_keep_just_here) is None
assert Matcher("outer").get_unique_op(z_keep_just_here) is None

# %%

NEW: NestEinsumsSpec = EinInfo((EinInfo((3, 2), "bd,ae"), 1), "bd,ae,bcd")

z = nest_einsums(x, (4, NEW, (EinInfo((0, 5), "abc,acd"), 6)))
assert_close(x.evaluate(), z_keep_just_here.evaluate())
print()
z_keep_just_here.print()

# %%

z_very_named = nest_einsums(
    x, (4, EinInfo((EinInfo((3, 2), "bd,ae"), 1), "bd,ae,bcd"), (EinInfo((0, 5), "abc,acd"), 6))
)
assert_close(x.evaluate(), z_very_named.evaluate())
print()
z_very_named.print()

# %%

z_very_named_rearrange = nest_einsums(z_very_named, (6, Rest(flat=True)))
assert_close(x.evaluate(), z_very_named_rearrange.evaluate())
print()
z_very_named_rearrange.print()

# %%

print()
term_early_out = nest_einsums(
    z_very_named,
    (3, (1, 4), Rest(flat=False)),
    traversal=new_traversal(term_early_at=Matcher.regex("^bd")),
)
assert_close(z_very_named.evaluate(), term_early_out.evaluate())
print()
term_early_out.print()

# %%

by_name = nest_einsums(x, ("eaf", ("ec", 2), Rest()))
assert_close(x.evaluate(), by_name.evaluate())
print()
by_name.print()

# %%

ignores_empty = nest_einsums(x, ("eaf", ("ec", 2), (), (), Rest()))
assert_close(x.evaluate(), ignores_empty.evaluate())
assert Matcher(lambda x: isinstance(x, Einsum) and len(x.args) == 0).get_unique_op(ignores_empty) is None
print()
ignores_empty.print()

# %%


fancy_permute = nest_einsums(x, ("eaf", ("ec", 2), EinInfo((3, 0), "hi"), Rest()))
assert_close(x.evaluate(), fancy_permute.evaluate())
print()
fancy_permute.print()

fancy_permute_2 = nest_einsums(x, ("eaf", ("ec", 2), EinInfo((3, 0), "hi", (1, 2, 0, 3)), Rest()))


assert_close(
    Matcher("hi").get_unique(fancy_permute).rearrange_str("a b c d -> b c a d").evaluate(),
    Matcher("hi").get_unique(fancy_permute_2).evaluate(),
)

# %%

empty_with_children = nest_einsums(
    Einsum.from_einsum_string(
        ",->",
        Einsum.from_einsum_string(
            "->",
        ),
        Einsum.from_einsum_string(
            "->",
        ),
    ),
    (),
)
empty_with_children.print()

# %%


sub_ein = Einsum.from_einsum_string(
    "eab,ae->ab", Array.randn(e, a, b, name="eab"), Array.randn(a, e, name="ae"), name="sub"
)
with_sub_ein = Einsum.from_einsum_string(
    "ab,ac->c",
    sub_ein,
    Array.randn(a, c, name="ac"),
)
unneeded_retain = nest_einsums(
    with_sub_ein,
    (2, (1, 0)),
)
assert_close(Matcher("sub").get_unique(unneeded_retain).evaluate(), sub_ein.evaluate())
unneeded_retain.print()

# %%

unneeded_retain_shrink = nest_einsums(
    with_sub_ein,
    (2, EinInfo((1, 0), shrink_out_axes=True, name="sub")),
)
assert_close(Matcher("sub").get_unique(unneeded_retain_shrink).evaluate(), sub_ein.evaluate().sum(dim=1))
unneeded_retain_shrink.print()

# %%

unneeded_retain_shrink_2 = nest_einsums(
    with_sub_ein,
    (2, EinInfo((1, 0), shrink_out_axes=True, name="sub")),
)
assert_close(Matcher("sub").get_unique(unneeded_retain_shrink_2).evaluate(), sub_ein.evaluate().sum(dim=1))
unneeded_retain_shrink_2.print()

# %%

z_new = nest_einsums(x, (("acd", ("bcd", Matcher.regex("^bd$"))), ("ae", "ec"), Rest()))
z_new.print()
sub_match_nest = nest_einsums(
    z_new,
    (4, 3, 0, 1, 2),
    IterativeMatcher.new_func(
        lambda _: IterateMatchResults(
            [new_traversal(term_early_at=Matcher.regex(r"^\w+ \* \w+$")), FINISHED, True],
            True,
        )
    ),
)
assert sub_match_nest == nest_einsums(
    z_new, ("eaf", "abc", "acd", NestMatcher({"bcd", "bd"}, flat=True), NestMatcher({"ae", "ec"}, flat=True))
)
